// ------------------------------------------------------------
// -----                  R3BOnlineSpectra                -----
// -----          Created April 13th 2016 by M.Heil       -----
// ------------------------------------------------------------

/*
 * This task should fill histograms with detector variables which allow
 * to test the detectors online
 *
 */

#include "R3BOnlineSpectraLosTest.h"
#include "R3BLosCalData.h"
#include "R3BLosMappedData.h"

#include "R3BRoluCalData.h"
#include "R3BRoluMappedData.h"

#include "R3BSci8CalData.h"
#include "R3BSci8MappedData.h"

#include "R3BBeamMonitorMappedData.h"

#include "R3BTofdCalData.h"
#include "R3BTofdMappedData.h"

#include "R3BPaddleCalData.h"
#include "R3BPaddleTamexMappedData.h"

#include "R3BPspxCalData.h"
#include "R3BPspxHitData.h"
#include "R3BPspxMappedData.h"
#include "R3BPspxPrecalData.h"

#include "../sofia/sofdata/twimData/R3BSofTwimCalData.h"
#include "../sofia/sofdata/twimData/R3BSofTwimHitData.h"
#include "../sofia/sofdata/twimData/R3BSofTwimMappedData.h"

#include "R3BEventHeader.h"
#include "R3BSamplerMappedData.h"
#include "R3BTCalEngine.h"

#include "R3BBunchedFiberCalData.h"
#include "R3BBunchedFiberHitData.h"
#include "R3BBunchedFiberMappedData.h"

#include "FairLogger.h"
#include "FairRootManager.h"
#include "FairRunAna.h"
#include "FairRunOnline.h"
#include "FairRuntimeDb.h"
#include "TCanvas.h"
#include "TGaxis.h"
#include "TH1F.h"
#include "TH2F.h"
#include "THttpServer.h"

#include "TClonesArray.h"
#include "TMath.h"
#include <TRandom3.h>
#include <TRandomGen.h>
#include <algorithm>
#include <array>
#include <cstdlib>
#include <ctime>
#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>
#define IS_NAN(x) TMath::IsNaN(x)
using namespace std;

R3BOnlineSpectraLosTest::R3BOnlineSpectraLosTest()
    : FairTask("OnlineSpectraLosTest", 1)
    , fMappedItemsTwim(NULL)
    , fCalItemsTwim(NULL)
    , fHitItemsTwim(NULL)
    , fTrigger(-1)
    , fTpat(-1)
    , fNofPlanes(N_PLANE_MAX_TOFD)
    , fPaddlesPerPlane(N_PADDLE_MAX_TOFD)
    , fClockFreq(1. / VFTX_CLOCK_MHZ * 1000.)
    , fNEvents(0)
{
}

R3BOnlineSpectraLosTest::R3BOnlineSpectraLosTest(const char* name, Int_t iVerbose)
    : FairTask(name, iVerbose)
    , fMappedItemsTwim(NULL)
    , fCalItemsTwim(NULL)
    , fHitItemsTwim(NULL)
    , fTrigger(-1)
    , fTpat(-1)
    , fNofPlanes(N_PLANE_MAX_TOFD)
    , fPaddlesPerPlane(N_PADDLE_MAX_TOFD)
    , fClockFreq(1. / VFTX_CLOCK_MHZ * 1000.)
    , fNEvents(0)
{
}

R3BOnlineSpectraLosTest::~R3BOnlineSpectraLosTest()
{
    //	delete fhTpat ;
    //	delete fhTrigger;
    //	delete fh_SEETRAM;
    for (int i = 0; i < NOF_FIB_DET; i++)
    {
        delete fh_channels_Fib[i];
        delete fh_fibers_Fib[i];
        delete fh_mult_Fib[i];
        delete fh_time_Fib[i];
        delete fh_multihit_m_Fib[i];
        delete fh_multihit_s_Fib[i];
        delete fh_ToT_m_Fib[i];
        delete fh_ToT_s_Fib[i];
        delete fh_ToT_single_Fib[i];
        delete fh_channels_single_Fib[i];
    }
    /*
      LOG(INFO) << "R3BSofTwimOnlineSpectra::Delete instance";
       if (fMappedItemsTwim)
           delete fMappedItemsTwim;
       if (fCalItemsTwim)
           delete fCalItemsTwim;
       if (fHitItemsTwim)
           delete fHitItemsTwim;
    */
}

InitStatus R3BOnlineSpectraLosTest::Init()
{

    // Initialize random number:
    std::srand(std::time(0)); // use current time as seed for random generator

    LOG(INFO) << "R3BOnlineSpectraLosTest::Init ";

    // try to get a handle on the EventHeader. EventHeader may not be
    // present though and hence may be null. Take care when using.

    FairRootManager* mgr = FairRootManager::Instance();
    if (NULL == mgr)
        LOG(fatal) << "FairRootManager not found";

    header = (R3BEventHeader*)mgr->GetObject("R3BEventHeader");
    if (!header)
        header = (R3BEventHeader*)mgr->GetObject("EventHeader.");

    FairRunOnline* run = FairRunOnline::Instance();

    run->GetHttpServer()->Register("/Tasks", this);

    // Get objects for detectors on all levels

    assert(DET_MAX + 1 == sizeof(fDetectorNames) / sizeof(fDetectorNames[0]));
    printf("Have %d fiber detectors.\n", NOF_FIB_DET);
    for (int det = 0; det < DET_MAX; det++)
    {
        fMappedItems.push_back((TClonesArray*)mgr->GetObject(Form("%sMapped", fDetectorNames[det])));
        if (NULL == fMappedItems.at(det))
        {
            printf("Could not find mapped data for '%s'.\n", fDetectorNames[det]);
        }
        fCalItems.push_back((TClonesArray*)mgr->GetObject(Form("%sCal", fDetectorNames[det])));
        fHitItems.push_back((TClonesArray*)mgr->GetObject(Form("%sHit", fDetectorNames[det])));
    }

    // get access to mapped data of the TWIM
    fMappedItemsTwim = (TClonesArray*)mgr->GetObject("TwimMappedData");
    if (!fMappedItemsTwim)
    {
        LOG(WARNING) << "R3BOnlineSpectra: TwimMappedData not found"; // return kFATAL;
    }

    // get access to cal data of the TWIM
    fCalItemsTwim = (TClonesArray*)mgr->GetObject("TwimCalData");
    if (!fCalItemsTwim)
        LOG(WARNING) << "R3BOnlineSpectra: TwimCalData not found";

    // get access to hit data of the TWIM
    fHitItemsTwim = (TClonesArray*)mgr->GetObject("TwimHitData");
    if (!fHitItemsTwim)
        LOG(WARNING) << "R3BOnlineSpectra: TwimHitData not found";

    //------------------------------------------------------------------------
    // create histograms of all detectors
    //------------------------------------------------------------------------

    // Trigger and Tpat
    TCanvas* cTrigg = new TCanvas("Trigg", "Triggers", 10, 10, 650, 350);
    fhTpat = new TH1F("Tpat", "Tpat", 20, 0, 20);
    fhTpat->GetXaxis()->SetTitle("Tpat value");

    fhTrigger = new TH1F("Trigger", "Trigger all", 20, 0, 20);
    fhTrigger->GetXaxis()->SetTitle("Trigger value");

    cTrigg->Divide(2, 1);
    cTrigg->cd(1);
    gPad->SetLogy();
    fhTrigger->Draw();
    cTrigg->cd(2);
    gPad->SetLogy();
    fhTpat->Draw();
    cTrigg->cd(0);

    //------------------------------------------------------------------------
    // Rolu detector

    if (fMappedItems.at(DET_ROLU))
    {
        TCanvas* cRolu = new TCanvas("Rolu", "Rolu", 10, 10, 800, 400);

        fh_rolu_tot = new TH2F("Rolu_tot", "ROLU ToT", 10, 0, 10, 3000, 0, 300);
        fh_rolu_tot->GetXaxis()->SetTitle("Channel number");
        fh_rolu_tot->GetYaxis()->SetTitle("ToT / ns");

        fh_rolu_channels = new TH1F("Rolu_channels", "ROLU channels", 10, 0, 10);
        fh_rolu_channels->GetXaxis()->SetTitle("Channel number");
        fh_rolu_channels->GetYaxis()->SetTitle("Counts");

        cRolu->Divide(2, 1);
        cRolu->cd(1);
        gPad->SetLogy();
        fh_rolu_channels->Draw();
        cRolu->cd(2);
        gPad->SetLogz();
        fh_rolu_tot->Draw("colz");
        cRolu->cd(0);
        run->AddObject(cRolu);

        run->GetHttpServer()->RegisterCommand("Reset_ROLU", Form("/Tasks/%s/->Reset_ROLU_Histo()", GetName()));
    }
    //------------------------------------------------------------------------
    // Sci8 detector

    if (fMappedItems.at(DET_SCI8))
    {
        TCanvas* cSci8 = new TCanvas("Sci8", "SCI8", 10, 10, 850, 850);

        fh_sci8_channels = new TH1F("sci8_channels", "SCI8 channels", 4, 0., 4.);
        fh_sci8_channels->GetXaxis()->SetTitle("Channel number");
        fh_sci8_multihit = new TH1F("sci8_multihit", "SCI8 multihit", 20, 0., 20.);
        fh_sci8_multihit->GetXaxis()->SetTitle("Multihit");
        fh_sci8_multihitVFTX = new TH2F("sci8_vftx_multihit", "SCI8 VFTX multihit", 4, 0., 4., 20, 0., 20.);
        fh_sci8_multihitVFTX->GetYaxis()->SetTitle("Multihit ");
        fh_sci8_multihitVFTX->GetXaxis()->SetTitle("Channel number");
        fh_sci8_multihitLEAD = new TH2F("sci8_lead_multihit", "SCI8 TAMEX Leading multihit", 4, 0., 4., 20, 0., 20.);
        fh_sci8_multihitLEAD->GetYaxis()->SetTitle("Multihit ");
        fh_sci8_multihitLEAD->GetXaxis()->SetTitle("Channel number");
        fh_sci8_multihitTRAI = new TH2F("sci8_trai_multihit", "SCI8 TAMEX Trailing multihit", 4, 0., 4., 20, 0., 20.);
        fh_sci8_multihitTRAI->GetYaxis()->SetTitle("Multihit ");
        fh_sci8_multihitTRAI->GetXaxis()->SetTitle("Channel number");
        fh_sci8_dt_hits = new TH1F("sci8_dt_hits", "SCI8 dt multihits", 40000, -2000, 2000);
        fh_sci8_dt_hits->GetXaxis()->SetTitle("dt_VFTX between two hits / ns");
        fh_sci8_dt_hits_l = new TH1F("sci8_dt_hits_l", "SCI8 dt multihits leading", 40000, -2000, 2000);
        fh_sci8_dt_hits_l->GetXaxis()->SetTitle("dt_LEAD between two hits / ns");
        fh_sci8_dt_hits_t = new TH1F("sci8_dt_hits_t", "SCI8 dt multihits trailing", 40000, -2000, 2000);
        fh_sci8_dt_hits_t->GetXaxis()->SetTitle("dt_TRAIL between two hits / ns");
        fh_sci8_tres_MCFD = new TH1F("sci8_time_res_MCFD", "SCI8 MCFD Time resolution - raw", 10000, -10., 10.);
        fh_sci8_tres_MCFD->GetXaxis()->SetTitle("Time MCFD / ns");
        fh_sci8_tres_TAMEX = new TH1F("sci8_time_res_TAMEX", "SCI8 TAMEX Time resolution -raw ", 8000, -4., 4.);
        fh_sci8_tres_TAMEX->GetXaxis()->SetTitle("Time TAMEX / ns");
        fh_sci8_tot = new TH2F("sci8_tot", "SCI8 ToT vs PM", 4, 0, 4, 3000, 0., 300.);
        fh_sci8_tot->GetXaxis()->SetTitle("PMT number");
        fh_sci8_tot->GetYaxis()->SetTitle("ToT / ns");
        fh_sci8_tot_mean = new TH1F("sci8_tot_mean", "SCI8 mean ToT", 30000, 0., 300.);
        fh_sci8_tot_mean->GetYaxis()->SetTitle("Counts");
        fh_sci8_tot_mean->GetXaxis()->SetTitle("ToT / ns");
        fh_tof_sci8 = new TH1F("sci8_tof", "SCI8-LOS ToF", 40000, -200, 200);
        fh_tof_sci8->GetXaxis()->SetTitle("ToF / ns");

        cSci8->Divide(2, 2);
        cSci8->cd(1);
        fh_sci8_channels->Draw();
        cSci8->cd(2);
        gPad->SetLogy();
        fh_sci8_multihit->Draw();
        // cSci8->cd(3);gPad->SetLogz();
        // fh_sci8_tot->Draw("colz");
        cSci8->cd(3);
        fh_sci8_tres_MCFD->Draw();
        // cSci8->cd(5);
        // fh_sci8_tres_TAMEX->Draw();
        cSci8->cd(4);
        fh_tof_sci8->Draw();
        // cSci8->cd(7);
        // fh_sci8_tot_mean->Draw();

        cSci8->cd(0);
        run->AddObject(cSci8);

        run->GetHttpServer()->RegisterCommand("Reset_SCI8", Form("/Tasks/%s/->Reset_SCI8_Histo()", GetName()));
    }

    if (fMappedItems.at(DET_BMON))
    {

        // get the theoretical calib factors for SEETRAM
        Double_t fexp = float(fsens_SEE + 9);
        Double_t fpow = float(pow(10., fexp));
        calib_SEE = 135641.7786 * fpow;
        cout << fsens_SEE << ", " << fexp << ", " << fpow << ", " << calib_SEE << endl;

        TCanvas* cbmon = new TCanvas("Beam_Monitor", "Beam Monitors", 820, 10, 900, 900);
        Int_t Nbin_bmon = reset_time / read_time;

        fhTpat = new TH1F("Tpat", "Tpat", 20, 0, 20);
        fhTpat->GetXaxis()->SetTitle("Tpat value");

        fhTrigger = new TH1F("Trigger", "Trigger all", 20, 0, 20);
        fhTrigger->GetXaxis()->SetTitle("Trigger value");

        fh_spill_length = new TH1F("spill_length", "Spill ", Nbin_bmon, 0, reset_time);
        fh_spill_length->GetXaxis()->SetTitle("time / sec");

        fh_IC = new TH1F("IC", "IC ", Nbin_bmon, 0, reset_time);
        fh_IC->GetXaxis()->SetTitle("time / sec");
        fh_IC->GetYaxis()->SetTitle("IC counts");

        fh_SEE = new TH1F("SEETRAM", "SEETRAM ", Nbin_bmon, 0, reset_time);
        fh_SEE->GetXaxis()->SetTitle("time / sec");
        fh_SEE->GetYaxis()->SetTitle("SEETRAM counts");

        fh_TOFDOR = new TH1F("TOFDOR", "TOFDOR ", Nbin_bmon, 0, reset_time);
        fh_TOFDOR->GetXaxis()->SetTitle("time / sec");
        fh_TOFDOR->GetYaxis()->SetTitle("TOFDOR counts");

        fh_IC_spill = new TH1F("IC_spill", "IC rate in kHz ", Nbin_bmon, 0, reset_time);
        fh_IC_spill->GetXaxis()->SetTitle("time / sec");
        fh_IC_spill->GetYaxis()->SetTitle("IC rate / kHz");

        fh_SEE_spill = new TH1F("SEE_spill", "SEE particle rate ", Nbin_bmon, 0, reset_time);
        fh_SEE_spill->GetXaxis()->SetTitle("time / sec");
        fh_SEE_spill->GetYaxis()->SetTitle("Particles / sec");

        fh_TOFDOR_spill = new TH1F("TOFDOR_spill", "TOFDOR rate in kHz ", Nbin_bmon, 0, reset_time);
        fh_TOFDOR_spill->GetXaxis()->SetTitle("time / sec");
        fh_TOFDOR_spill->GetYaxis()->SetTitle("TOFDOR rate / kHz");

        cbmon->Divide(3, 3);
        cbmon->cd(1);
        gPad->SetLogy();
        fhTrigger->Draw();
        cbmon->cd(2);
        gPad->SetLogy();
        fhTpat->Draw();
        cbmon->cd(3);
        fh_spill_length->Draw();
        cbmon->cd(4);
        fh_IC->Draw("hist");
        //        fh_TimePreviousEvent->Draw("hist");
        cbmon->cd(5);
        fh_SEE->Draw("hist");
        cbmon->cd(6);
        fh_TOFDOR->Draw("hist");
        cbmon->cd(7);
        fh_IC_spill->Draw("hist");
        cbmon->cd(8);
        fh_SEE_spill->Draw("hist");
        cbmon->cd(9);
        fh_TOFDOR_spill->Draw("hist");
        cbmon->cd(0);

        run->AddObject(cbmon);

        run->GetHttpServer()->RegisterCommand("Reset_BMON", Form("/Tasks/%s/->Reset_BMON_Histo()", GetName()));
    }

    //------------------------------------------------------------------------
    // Los detector

    TCanvas* cLos[fNofLosDetectors];

    if (fMappedItems.at(DET_LOS))
    {

        if (fNofLosDetectors == 2)
        {
            fh_los_ToT_vs_ToT = new TH2F("los_tot_vs_tot", "LOS ToT vs ToT", 3000, 0., 300., 3000, 0., 300.);
            fh_los_ToT_vs_ToT->GetXaxis()->SetTitle("ToT LOS1 / ns");
            fh_los_ToT_vs_ToT->GetYaxis()->SetTitle("ToT LOS2 / ns");
            fh_los_nPart_vs_nPart = new TH2F("los_nPart_vs_nPart", "LOS nPart vs nPart", 20, 0., 20., 20, 0., 20.);
            fh_los_nPart_vs_nPart->GetXaxis()->SetTitle("nPart LOS1");
            fh_los_nPart_vs_nPart->GetYaxis()->SetTitle("nPart LOS2");
            fh_dt_los1_los2 = new TH1F("los_dt_los1_los2", "LOS dt los1 los2", 20000, -10, 10);
            fh_dt_los1_los2->GetXaxis()->SetTitle("dt / ns");
            fh_dt_los1_los2->GetYaxis()->SetTitle("Counts");

            for (Int_t i = 0; i < 8; i++)
            {
                char strName1[255];
                sprintf(strName1, "test_PM%d", i + 1);
                char strName2[255];
                sprintf(strName2, "test PM%d", i + 1);
                fh_test[i] = new TH2F(strName1, strName2, 1000, 0, 10, 400, 1, 400.);
                char strName3[255];
                sprintf(strName3, "test_corr_PM%d", i + 1);
                char strName4[255];
                sprintf(strName4, "test_corr PM%d", i + 1);
                fh_test_corr[i] = new TH2F(strName1, strName2, 1000, 0, 10, 400, 1, 400.);
            }

            TCanvas* cLosCor = new TCanvas("Los_correl", "Los correl", 10, 10, 910, 810);
            cLosCor->Divide(4, 5);
            cLosCor->cd(1);
            gPad->SetLogz();
            fh_los_ToT_vs_ToT->Draw("colz");
            cLosCor->cd(2);
            gPad->SetLogz();
            fh_los_nPart_vs_nPart->Draw("colz");
            cLosCor->cd(3);
            gPad->SetLogy();
            fh_dt_los1_los2->Draw();
            cLosCor->cd(4);
            gPad->SetLogz();
            fh_test[0]->Draw("colz");
            cLosCor->cd(5);
            gPad->SetLogz();
            fh_test[1]->Draw("colz");
            cLosCor->cd(6);
            gPad->SetLogz();
            fh_test[2]->Draw("colz");
            cLosCor->cd(7);
            gPad->SetLogz();
            fh_test[3]->Draw("colz");
            cLosCor->cd(8);
            gPad->SetLogz();
            fh_test[4]->Draw("colz");
            cLosCor->cd(9);
            gPad->SetLogz();
            fh_test[5]->Draw("colz");
            cLosCor->cd(10);
            gPad->SetLogz();
            fh_test[6]->Draw("colz");
            cLosCor->cd(11);
            gPad->SetLogz();
            fh_test[7]->Draw("colz");
            cLosCor->cd(13);
            gPad->SetLogz();
            fh_test_corr[0]->Draw("colz");
            cLosCor->cd(14);
            gPad->SetLogz();
            fh_test_corr[1]->Draw("colz");
            cLosCor->cd(15);
            gPad->SetLogz();
            fh_test_corr[2]->Draw("colz");
            cLosCor->cd(16);
            gPad->SetLogz();
            fh_test_corr[3]->Draw("colz");
            cLosCor->cd(17);
            gPad->SetLogz();
            fh_test_corr[4]->Draw("colz");
            cLosCor->cd(18);
            gPad->SetLogz();
            fh_test_corr[5]->Draw("colz");
            cLosCor->cd(19);
            gPad->SetLogz();
            fh_test_corr[6]->Draw("colz");
            cLosCor->cd(20);
            gPad->SetLogz();
            fh_test_corr[7]->Draw("colz");
        }

        //	fh_los_tres_MTDC = new TH1F("los_time_res_MTDC", "LOS MTDC Time resolution - raw", 4000, -4, 4);
        //	fh_los_tres_MTDC->GetXaxis()->SetTitle("Time MTDC / ns");

        for (Int_t iloscount = 0; iloscount < fNofLosDetectors; iloscount++)
        {

            char detName[255];
            sprintf(detName, "LOS%d", iloscount + 1);

            cLos[iloscount] = new TCanvas(detName, detName, 10, 10, 1010, 810);

            fh_los_twin[iloscount] = new TH2F(
                Form("%s_ToT_TWIN", detName), Form("%s ToT vs dE Twin ", detName), 800, 0, 400., 10000, 0., 5000.);
            fh_los_twin[iloscount]->GetXaxis()->SetTitle("LOS ToT / ns");
            fh_los_twin[iloscount]->GetYaxis()->SetTitle("dE TWIN / channel");

            fh_los_channels[iloscount] =
                new TH1F(Form("%s_channels", detName), Form("%s channels", detName), 10, 0., 10.);
            fh_los_channels[iloscount]->GetXaxis()->SetTitle("Channel number");

            fh_los_multihit[iloscount] =
                new TH1F(Form("%s_multihit", detName), Form("%s multihit && all 8 PMs", detName), 30, 0., 30.);
            fh_los_multihit[iloscount]->GetXaxis()->SetTitle("Multihit");

            fh_los_pos_MCFD[iloscount] =
                new TH2F(Form("%s_pos_MCFD", detName), Form("%s MCFD Position ", detName), 500, -5., 5., 500, -5., 5.);
            fh_los_pos_MCFD[iloscount]->GetXaxis()->SetTitle("X position / cm");
            fh_los_pos_MCFD[iloscount]->GetYaxis()->SetTitle("Y position / cm");

            fh_los_pos_TAMEX[iloscount] = new TH2F(
                Form("%s_pos_TAMEX", detName), Form("%s TAMEX Position ", detName), 500, -5., 5., 500, -5., 5.);
            fh_los_pos_TAMEX[iloscount]->GetXaxis()->SetTitle("X position / cm");
            fh_los_pos_TAMEX[iloscount]->GetYaxis()->SetTitle("Y position / cm");

            fh_los_pos_ToT[iloscount] =
                new TH2F(Form("%s_pos_ToT", detName), Form("%s ToT Position ", detName), 500, -5., 5., 500, -5., 5.);
            fh_los_pos_ToT[iloscount]->GetXaxis()->SetTitle("X position / cm");
            fh_los_pos_ToT[iloscount]->GetYaxis()->SetTitle("Y position / cm");

            fh_los_dt_hits_ToT[iloscount] =
                new TH2F(Form("%s_dt_ToT", detName), Form("%s ToT dt ", detName), 1000, 0, 10, 300, 0, 300.);
            fh_los_dt_hits_ToT[iloscount]->GetXaxis()->SetTitle(
                "dt between two hits / micros");                             // dt between two hits / ns
            fh_los_dt_hits_ToT[iloscount]->GetYaxis()->SetTitle("ToT / ns"); // ToT / ns

            fh_los_tres_MCFD[iloscount] = new TH1F(
                Form("%s_time_res_MCFD", detName), Form("%s MCFD Time resolution - raw", detName), 4000, -4., 4.);
            fh_los_tres_MCFD[iloscount]->GetXaxis()->SetTitle("Time MCFD / ns");

            fh_los_tres_TAMEX[iloscount] = new TH1F(
                Form("%s_time_res_TAMEX", detName), Form("%s TAMEX Time resolution -raw ", detName), 4000, -4., 4.);
            fh_los_tres_TAMEX[iloscount]->GetXaxis()->SetTitle("Time TAMEX / ns");

            fh_los_tot[iloscount] =
                new TH2F(Form("%s_tot", detName), Form("%s ToT vs PMT", detName), 10, 0, 10, 1500, 0., 300.);
            fh_los_tot[iloscount]->GetXaxis()->SetTitle("PMT number");
            fh_los_tot[iloscount]->GetYaxis()->SetTitle("ToT / ns");

            fh_los_tot_mean[iloscount] =
                new TH1F(Form("%s_tot_mean", detName), Form("%s mean ToT", detName), 1500, 0., 300.);
            fh_los_tot_mean[iloscount]->GetYaxis()->SetTitle("Counts");
            fh_los_tot_mean[iloscount]->GetXaxis()->SetTitle("ToT / ns");

            fh_los_tot_mean_corr[iloscount] =
                new TH1F(Form("%s_tot_mean_cor", detName), Form("%s mean ToT_cor", detName), 2000, 1., 400.);
            fh_los_tot_mean_corr[iloscount]->GetYaxis()->SetTitle("Counts");
            fh_los_tot_mean_corr[iloscount]->GetXaxis()->SetTitle("ToT_cor / ns");

            fh_los_ihit_ToT[iloscount] =
                new TH2F(Form("%s_tot_ihit", detName), Form("%s ToT vs ihit", detName), 1000, 0, 10, 600, 0., 300.);
            // fh_los_ihit_ToT[iloscount]= new TH2F(Form("%s_tot_ihit",detName),Form("%s ToT vs
            // ihit",detName),15,0,15,300,0.,300.);
            fh_los_ihit_ToT[iloscount]->GetXaxis()->SetTitle("iHit");
            fh_los_ihit_ToT[iloscount]->GetYaxis()->SetTitle("ToT / ns");

            fh_losToT_vs_Events[iloscount] = new TH2F(Form("%s_tot_vs_event", detName),
                                                      Form("%s ToT vs. Event #", detName),
                                                      10000,
                                                      0,
                                                      10000000,
                                                      300,
                                                      0.,
                                                      300.);
            fh_losToT_vs_Events[iloscount]->GetYaxis()->SetTitle("ToT / ns");
            fh_losToT_vs_Events[iloscount]->GetXaxis()->SetTitle("Event number");

            fh_losTAMEX_vs_Events[iloscount] = new TH2F(Form("%s_dtTAMEX_vs_event", detName),
                                                        Form("%s dtTAMEX vs. Event #", detName),
                                                        10000,
                                                        0,
                                                        10000000,
                                                        1000,
                                                        -4.,
                                                        4.);
            fh_losTAMEX_vs_Events[iloscount]->GetYaxis()->SetTitle("Time TAMEX  / ns");
            fh_losTAMEX_vs_Events[iloscount]->GetXaxis()->SetTitle("Event number");

            fh_losMCFD_vs_Events[iloscount] = new TH2F(Form("%s_dtMCFD_vs_event", detName),
                                                       Form("%s dtMCFD vs. Event #", detName),
                                                       10000,
                                                       0,
                                                       10000000,
                                                       1000,
                                                       -4.,
                                                       4.);
            fh_losMCFD_vs_Events[iloscount]->GetYaxis()->SetTitle("Time MCFD  / ns");
            fh_losMCFD_vs_Events[iloscount]->GetXaxis()->SetTitle("Event number");

            fh_los_rw[iloscount] = new TH1F(Form("%s_rw", detName), Form("%s rw", detName), 600, 100, 400.);
            fh_los_rw[iloscount]->GetYaxis()->SetTitle("Counts");
            fh_los_rw[iloscount]->GetXaxis()->SetTitle("Emission angle / deg");

            fh_los_vftx_tamex[iloscount] =
                new TH1F(Form("%s_vftx_tamex", detName), Form("%s vftx_tamex", detName), 100000, -5000, 5000.);
            fh_los_vftx_tamex[iloscount]->GetYaxis()->SetTitle("Counts");
            fh_los_vftx_tamex[iloscount]->GetXaxis()->SetTitle("Ttamex-Tvftx / ns");

            fh_los_tot_S8[iloscount] = new TH2F(
                Form("%s_tot_dtS8", detName), Form("%s ToTcorr vs. dT-S8", detName), 2000, -5, 5, 400, 1., 401.);

            cLos[iloscount]->Divide(5, 3);
            cLos[iloscount]->cd(1);
            fh_los_channels[iloscount]->Draw();
            cLos[iloscount]->cd(2);
            gPad->SetLogy();
            fh_los_multihit[iloscount]->Draw();
            cLos[iloscount]->cd(3);
            gPad->SetLogz();
            fh_los_tot[iloscount]->Draw("colz");
            cLos[iloscount]->cd(4); // gPad->SetLogy();
            fh_los_tot_mean[iloscount]->Draw();
            cLos[iloscount]->cd(5); // gPad->SetLogy();
            fh_los_tot_mean_corr[iloscount]->Draw();
            cLos[iloscount]->cd(6);
            gPad->SetLogy();
            fh_los_tres_MCFD[iloscount]->Draw();
            cLos[iloscount]->cd(7);
            gPad->SetLogy();
            fh_los_tres_TAMEX[iloscount]->Draw();
            cLos[iloscount]->cd(8);
            gPad->SetLogy();
            fh_los_vftx_tamex[iloscount]->Draw();
            cLos[iloscount]->cd(9);
            gPad->SetLogz();
            fh_los_tot_S8[iloscount]->Draw("colz");
            cLos[iloscount]->cd(10);
            gPad->SetLogz();
            fh_los_twin[iloscount]->Draw("colz");
            cLos[iloscount]->cd(11);
            gPad->SetLogz();
            fh_los_pos_ToT[iloscount]->Draw("colz");
            cLos[iloscount]->cd(12);
            gPad->SetLogz();
            fh_los_pos_MCFD[iloscount]->Draw("colz");
            cLos[iloscount]->cd(13);
            gPad->SetLogz();
            fh_los_pos_TAMEX[iloscount]->Draw("colz");
            cLos[iloscount]->cd(14);
            gPad->SetLogz();
            fh_los_ihit_ToT[iloscount]->Draw("colz");
            cLos[iloscount]->cd(15); // gPad->SetLogy();
            fh_los_rw[iloscount]->Draw();
            cLos[iloscount]->cd(0);
            run->AddObject(cLos[iloscount]);
        }

        run->GetHttpServer()->RegisterCommand("Reset_LOS", Form("/Tasks/%s/->Reset_LOS_Histo()", GetName()));
    }
    if (fMappedItemsTwim)
    {
        // TWIM: Map data for E in the first (anodes from 0 to 7) and final sections (anodes from 8 to 15)
        TCanvas* cTwimMap_ESum = new TCanvas("twim_ESum", "twim ESum", 10, 10, 800, 700);
        cTwimMap_ESum->Divide(3, 2);

        cTwimMap_ESum->cd(1);
        gPad->SetLogy();
        fh1_twim_ESum[0] = new TH1F("fh1_twim_ESum1", "twim:ESum:first", 8192, 0, 8192);
        fh1_twim_ESum[0]->GetXaxis()->SetTitle("Energy [channels]");
        fh1_twim_ESum[0]->GetYaxis()->SetTitle("Counts");
        fh1_twim_ESum[0]->GetYaxis()->SetTitleOffset(1.1);
        fh1_twim_ESum[0]->GetXaxis()->CenterTitle(true);
        fh1_twim_ESum[0]->GetYaxis()->CenterTitle(true);
        fh1_twim_ESum[0]->GetXaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[0]->GetXaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[0]->GetYaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[0]->GetYaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[0]->Draw("");
        cTwimMap_ESum->cd(2);
        gPad->SetLogy();
        fh1_twim_ESum[1] = new TH1F("fh1_twim_ESum2", "twim:ESum:Second", 8192, 0, 8192);
        fh1_twim_ESum[1]->GetXaxis()->SetTitle("Energy [channels]");
        fh1_twim_ESum[1]->GetYaxis()->SetTitle("Counts");
        fh1_twim_ESum[1]->GetYaxis()->SetTitleOffset(1.1);
        fh1_twim_ESum[1]->GetXaxis()->CenterTitle(true);
        fh1_twim_ESum[1]->GetYaxis()->CenterTitle(true);
        fh1_twim_ESum[1]->GetXaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[1]->GetXaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[1]->GetYaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[1]->GetYaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[1]->Draw("");
        cTwimMap_ESum->cd(3);
        gPad->SetLogy();
        fh1_twim_ESum[2] = new TH1F("fh1_twim_ESum", "twim:ESum", 8192, 0, 8192);
        fh1_twim_ESum[2]->GetXaxis()->SetTitle("Energy [channels]");
        fh1_twim_ESum[2]->GetYaxis()->SetTitle("Counts");
        fh1_twim_ESum[2]->GetYaxis()->SetTitleOffset(1.1);
        fh1_twim_ESum[2]->GetXaxis()->CenterTitle(true);
        fh1_twim_ESum[2]->GetYaxis()->CenterTitle(true);
        fh1_twim_ESum[2]->GetXaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[2]->GetXaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[2]->GetYaxis()->SetLabelSize(0.045);
        fh1_twim_ESum[2]->GetYaxis()->SetTitleSize(0.045);
        fh1_twim_ESum[2]->Draw("");
        cTwimMap_ESum->cd(4);
        gPad->SetLogz();
        fh2_twim_ESum = new TH2F("fh2_twim_ESum", "twim: ESum1 vs Esum2", 2192, 0, 8192, 2192, 0, 8192);
        fh2_twim_ESum->GetXaxis()->SetTitle("Energy1 [channels]");
        fh2_twim_ESum->GetYaxis()->SetTitle("Energy2 [channels]");
        fh2_twim_ESum->GetYaxis()->SetTitleOffset(1.1);
        fh2_twim_ESum->GetXaxis()->CenterTitle(true);
        fh2_twim_ESum->GetYaxis()->CenterTitle(true);
        fh2_twim_ESum->GetXaxis()->SetLabelSize(0.045);
        fh2_twim_ESum->GetXaxis()->SetTitleSize(0.045);
        fh2_twim_ESum->GetYaxis()->SetLabelSize(0.045);
        fh2_twim_ESum->GetYaxis()->SetTitleSize(0.045);
        fh2_twim_ESum->Draw("col");
        cTwimMap_ESum->cd(5);
        fh1_Twimmap_mult = new TH1F("twin_multipl", "Twin multiplicity", 17, 0, 17);
        fh1_Twimmap_mult->GetXaxis()->SetTitle("Anode");
        fh1_Twimmap_mult->GetYaxis()->SetTitle("Counts");
        fh1_Twimmap_mult->GetYaxis()->SetTitleOffset(1.1);
        fh1_Twimmap_mult->GetXaxis()->CenterTitle(true);
        fh1_Twimmap_mult->GetYaxis()->CenterTitle(true);
        fh1_Twimmap_mult->GetXaxis()->SetLabelSize(0.045);
        fh1_Twimmap_mult->GetXaxis()->SetTitleSize(0.045);
        fh1_Twimmap_mult->GetYaxis()->SetLabelSize(0.045);
        fh1_Twimmap_mult->GetYaxis()->SetTitleSize(0.045);
        fh1_Twimmap_mult->Draw("");

        run->AddObject(cTwimMap_ESum);

        run->GetHttpServer()->RegisterCommand("Reset_TWIN", Form("/Tasks/%s/->Reset_TWIN_Histo()", GetName()));
    }

    //-----------------------------------------------------------------------
    // Fiber Detectors 1-NOF_FIB_DET

    char canvName[255];
    TCanvas* FibCanvas[NOF_FIB_DET];

    for (Int_t ifibcount = 0; ifibcount < NOF_FIB_DET; ifibcount++)
    {

        std::stringstream fibhitobj;
        std::stringstream fibmapobj;
        std::stringstream FiName;
        std::string temp;

        if (fMappedItems.at(DET_FI_FIRST + ifibcount))
        {

            const char* chistName;
            const char* chistTitle;
            const char* detName;
            std::string tempTitle;
            std::string tempName;
            std::stringstream tempFibName;
            std::string tempFibNames;
            std::stringstream tempCanvName;

            detName = fDetectorNames[DET_FI_FIRST + ifibcount];

            FibCanvas[ifibcount] = new TCanvas(detName, detName, 10, 10, 910, 910);

            // Channels:
            fh_channels_Fib[ifibcount] =
                new TH1F(Form("%s_channels", detName), Form("%s channels", detName), 520, 0., 520.);
            fh_channels_Fib[ifibcount]->GetXaxis()->SetTitle("Channel number");
            fh_channels_Fib[ifibcount]->GetYaxis()->SetTitle("Counts");

            // Channels:
            fh_channels_single_Fib[ifibcount] =
                new TH1F(Form("%s_channels_single", detName), Form("%s channels of single PMTs", detName), 10, 0., 10.);
            fh_channels_single_Fib[ifibcount]->GetXaxis()->SetTitle("Channel number");
            fh_channels_single_Fib[ifibcount]->GetYaxis()->SetTitle("Counts");

            // Fibers:
            fh_fibers_Fib[ifibcount] =
                new TH1F(Form("%s_fibers", detName), Form("%s fibers", detName), N_FIBER_PLOT, 0., N_FIBER_PLOT);
            fh_fibers_Fib[ifibcount]->GetXaxis()->SetTitle("Fiber number");
            fh_fibers_Fib[ifibcount]->GetYaxis()->SetTitle("Counts");

            // Multiplicity (number of hit fibers):
            fh_mult_Fib[ifibcount] = new TH1F(Form("%s_mult", detName), Form("%s mult", detName), 100, 0., 100.);
            fh_mult_Fib[ifibcount]->GetXaxis()->SetTitle("Multiplicity");
            fh_mult_Fib[ifibcount]->GetYaxis()->SetTitle("Counts");

            // Multihit MAPMT:
            fh_multihit_m_Fib[ifibcount] = new TH2F(
                Form("%s_multihit_m", detName), Form("%s multihits MAPMT", detName), 520, 0., 520., 20, 0., 20.);
            fh_multihit_m_Fib[ifibcount]->GetXaxis()->SetTitle("MAPMT channel");
            fh_multihit_m_Fib[ifibcount]->GetYaxis()->SetTitle("Multihit");

            // Multihit SAPMT:
            fh_multihit_s_Fib[ifibcount] = new TH2F(
                Form("%s_multihit_s", detName), Form("%s multihits single PMT", detName), 16, 0., 16., 20, 0., 20.);
            fh_multihit_s_Fib[ifibcount]->GetXaxis()->SetTitle("Single PMT channel");
            fh_multihit_s_Fib[ifibcount]->GetYaxis()->SetTitle("Multihit");

            // ToT MAPMT:
            fh_ToT_m_Fib[ifibcount] = new TH2F(Form("%s_tot_m", detName),
                                               Form("%s ToT of MAPMT", detName),
                                               N_FIBER_PLOT,
                                               0.,
                                               N_FIBER_PLOT,
                                               600,
                                               0.,
                                               200.);
            fh_ToT_m_Fib[ifibcount]->GetXaxis()->SetTitle("Fiber number");
            fh_ToT_m_Fib[ifibcount]->GetYaxis()->SetTitle("ToT / ns");

            // ToT SAPMT:
            fh_ToT_s_Fib[ifibcount] = new TH2F(Form("%s_tot_s", detName),
                                               Form("%s ToT of single PMT", detName),
                                               N_FIBER_PLOT,
                                               0.,
                                               N_FIBER_PLOT,
                                               600,
                                               0.,
                                               200.);
            fh_ToT_s_Fib[ifibcount]->GetXaxis()->SetTitle("Fiber number");
            fh_ToT_s_Fib[ifibcount]->GetYaxis()->SetTitle("ToT / ns");

            // Time of fiber:
            fh_time_Fib[ifibcount] = new TH2F(Form("%s_TimevsFiber", detName),
                                              Form("%s Time vs Fiber", detName),
                                              N_FIBER_PLOT,
                                              0.,
                                              N_FIBER_PLOT,
                                              2048,
                                              -1024.,
                                              1024.);
            fh_time_Fib[ifibcount]->GetXaxis()->SetTitle("Fiber number");
            fh_time_Fib[ifibcount]->GetYaxis()->SetTitle("tMAPMT-tSPMT");

            // ToF LOS -> Fiber:
            fh_Fib_ToF[ifibcount] = new TH2F(Form("%s_tof", detName),
                                             Form("%s ToF LOS to Fiber", detName),
                                             N_FIBER_PLOT,
                                             0.,
                                             N_FIBER_PLOT,
                                             10000,
                                             -5000.,
                                             5000.);
            fh_Fib_ToF[ifibcount]->GetYaxis()->SetTitle("ToF / ns");
            fh_Fib_ToF[ifibcount]->GetXaxis()->SetTitle("Fiber number");

            // Not-calibrated position:
            fh_Fib_pos[ifibcount] =
                new TH1F(Form("%s_pos", detName), Form("%s Not-calibrated position", detName), 2100, -1050., 1050.);
            fh_Fib_pos[ifibcount]->GetXaxis()->SetTitle("Position");
            fh_Fib_pos[ifibcount]->GetYaxis()->SetTitle("Counts");

            // hit fiber number vs. event number:
            fh_Fib_vs_Events[ifibcount] = new TH2F(Form("%s_fib_vs_event", detName),
                                                   Form("%s Fiber # vs. Event #", detName),
                                                   10000,
                                                   0,
                                                   5e6,
                                                   1100,
                                                   0.,
                                                   1100.);
            fh_Fib_vs_Events[ifibcount]->GetYaxis()->SetTitle("Fiber number");
            fh_Fib_vs_Events[ifibcount]->GetXaxis()->SetTitle("Event number");

            // ToT SAPMT:
            fh_ToT_single_Fib[ifibcount] = new TH2F(
                Form("%s_tot_single", detName), Form("%s ToT of single PMTs", detName), 10, 0., 10., 600, 0., 200.);
            fh_ToT_single_Fib[ifibcount]->GetXaxis()->SetTitle("Fiber number");
            fh_ToT_single_Fib[ifibcount]->GetYaxis()->SetTitle("ToT / ns");

            FibCanvas[ifibcount]->Divide(3, 4);
            FibCanvas[ifibcount]->cd(1);
            fh_channels_Fib[ifibcount]->Draw();
            FibCanvas[ifibcount]->cd(2);
            gPad->SetLogz();
            fh_multihit_m_Fib[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(3);
            gPad->SetLogz();
            fh_multihit_s_Fib[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(4);
            fh_fibers_Fib[ifibcount]->Draw();
            FibCanvas[ifibcount]->cd(5);
            fh_mult_Fib[ifibcount]->Draw();
            FibCanvas[ifibcount]->cd(6);
            gPad->SetLogz();
            fh_ToT_m_Fib[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(7);
            gPad->SetLogz();
            fh_ToT_s_Fib[ifibcount]->Draw("colz");
            // FibCanvas[ifibcount]->cd(8);gPad->SetLogz();
            // fh_time_Fib[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(8);
            gPad->SetLogz();
            fh_Fib_ToF[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(9);
            fh_Fib_pos[ifibcount]->Draw();
            FibCanvas[ifibcount]->cd(10);
            fh_Fib_vs_Events[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(11);
            fh_ToT_single_Fib[ifibcount]->Draw("colz");
            FibCanvas[ifibcount]->cd(12);
            fh_channels_single_Fib[ifibcount]->Draw();

            FibCanvas[ifibcount]->cd(0);
            run->AddObject(FibCanvas[ifibcount]);

        } // end if(Mapped)

    } // end for(ifibcount)
    if (NOF_FIB_DET > 0)
        run->GetHttpServer()->RegisterCommand("Reset_Fibers", Form("/Tasks/%s/->Reset_FIBERS_Histo()", GetName()));

    //---------------------------------------------------------------------------------------
    // Ptof detector

    if (fCalItems.at(DET_PTOF))
    {

        TCanvas* cPtof_plane = new TCanvas("Ptof_plane", "Ptof plane", 10, 10, 510, 510);
        cPtof_plane->Divide(1, 2);

        fh_ptof_channels = new TH1F("Ptof_channels", "Ptof channels", 65, 0., 65.);
        cPtof_plane->cd(1);
        fh_ptof_channels->Draw();

        fh_ptof_channels_cut = new TH1F("Ptof_channels_cut", "Ptof channels with cut", 65, 0., 65.);
        cPtof_plane->cd(2);
        fh_ptof_channels_cut->Draw();
        run->AddObject(cPtof_plane);

        TCanvas* cPtof_test = new TCanvas("Ptof_test", "Ptof test", 10, 10, 510, 510);
        cPtof_test->Divide(1, 1);

        fh_ptof_test1 = new TH1F("Ptof_test1", "Ptof test1", 1000, 0., 100.);
        fh_ptof_test2 = new TH1F("Ptof_test2", "Ptof test2", 1000, 0., 100.);
        cPtof_test->cd(1);
        fh_ptof_test1->Draw();
        //		cPtof_test->cd(2);
        //		fh_ptof_test2->Draw();

        run->AddObject(cPtof_test);

        TCanvas* cPtof1;
        cPtof1 = new TCanvas("Ptof1", "Ptof1", 10, 10, 500, 500);
        cPtof1->Divide(3, 3);
        for (Int_t j = 1; j < 10; j++)
        {
            char strName[255];
            sprintf(strName, "ToT_Bar_%d_PM_1", j);
            fh_ptof_TotPm1[j] = new TH1F(strName, "", 1000, 0., 100.);
            cPtof1->cd(j);
            fh_ptof_TotPm1[j]->Draw();
        }
        run->AddObject(cPtof1);

        TCanvas* cPtof2;
        cPtof2 = new TCanvas("Ptof2", "Ptof2", 10, 10, 500, 500);
        cPtof2->Divide(3, 3);
        for (Int_t j = 1; j < 10; j++)
        {
            char strName[255];
            sprintf(strName, "ToT_Bar_%d_PM_2", j);
            fh_ptof_TotPm2[j] = new TH1F(strName, "", 1000, 0., 100.);
            cPtof2->cd(j);
            fh_ptof_TotPm2[j]->Draw();
        }
        run->AddObject(cPtof2);
    }

    //---------------------------------------------------------------------------------------------------
    // TofD detector

    if (fMappedItems.at(DET_TOFD))
    {
        TCanvas* cTofd_planes = new TCanvas("TOFD_planes", "TOFD planes", 10, 10, 1100, 1000);
        cTofd_planes->Divide(5, 4);

        fh_TimePreviousEvent = new TH1F("TimePreviousEvent", "Time between 2 particles ", 300000, -3000, 3000);
        fh_TimePreviousEvent->GetXaxis()->SetTitle("time / µsec");
        fh_TimePreviousEvent->GetYaxis()->SetTitle("counts");

        for (Int_t j = 0; j < 4; j++)
        {
            char strName1[255];
            sprintf(strName1, "tofd_channels_plane_%d", j + 1);
            char strName2[255];
            sprintf(strName2, "Tofd channels plane %d", j + 1);
            fh_tofd_channels[j] = new TH1F(strName1, strName2, 90, -45., 45.);
            fh_tofd_channels[j]->GetXaxis()->SetTitle("Channel");
            fh_tofd_channels[j]->GetYaxis()->SetTitle("Counts");

            char strName3[255];
            sprintf(strName3, "tofd_ToT_plane_%d", j + 1);
            char strName4[255];
            sprintf(strName4, "Tofd ToT plane %d", j + 1);
            fh_tofd_TotPm[j] = new TH2F(strName3, strName4, 90, -45, 45, 3000, 0., 300.);
            fh_tofd_TotPm[j]->GetXaxis()->SetTitle("Bar number");
            fh_tofd_TotPm[j]->GetYaxis()->SetTitle("ToT / ns");

            char strName7[255];
            sprintf(strName7, "tofd_ToF_plane_%d", j + 1);
            char strName8[255];
            sprintf(strName8, "TOFD-LOS ToF plane %d", j + 1);
            fh_tofd_ToF[j] = new TH2F(strName7, strName8, 15, 0, 15, 1000, 0, 100);
            fh_tofd_ToF[j]->GetXaxis()->SetTitle("Bar number");
            fh_tofd_ToF[j]->GetYaxis()->SetTitle("ToF / ns");

            char strName9[255];
            sprintf(strName9, "tofd_multihit_plane_%d", j + 1);
            char strName10[255];
            sprintf(strName10, "Tofd multihit plane %d", j + 1);
            fh_tofd_multihit[j] = new TH2F(strName9, strName10, 45, 0., 45., 10, 0, 10);
            fh_tofd_multihit[j]->GetXaxis()->SetTitle("Bar number");
            fh_tofd_multihit[j]->GetYaxis()->SetTitle("Multihit");

            if (j < 3)
            {
                int jk = j + 1;
                char strName5[255];
                sprintf(strName5, "tofd_ToF_plane_%d_%d", jk, jk + 1);
                char strName6[255];
                sprintf(strName6, "Tofd ToF plane %d and %d ", jk, jk + 1);
                fh_tofd_dt[j] = new TH2F(strName5, strName6, 50, 0, 50, 800, -40., 40);
                fh_tofd_dt[j]->GetXaxis()->SetTitle("Bar number");
                fh_tofd_dt[j]->GetYaxis()->SetTitle("dt / ns");
            }
        }

        cTofd_planes->cd(1);
        fh_tofd_channels[0]->Draw();
        cTofd_planes->cd(2);
        gPad->SetLogz();
        fh_tofd_TotPm[0]->Draw("colz");
        cTofd_planes->cd(3);
        gPad->SetLogz();
        fh_tofd_ToF[0]->Draw("colz");
        cTofd_planes->cd(4);
        gPad->SetLogz();
        fh_tofd_multihit[0]->Draw("colz");

        cTofd_planes->cd(6);
        fh_tofd_channels[1]->Draw();
        cTofd_planes->cd(7);
        gPad->SetLogz();
        fh_tofd_TotPm[1]->Draw("colz");
        cTofd_planes->cd(8);
        gPad->SetLogz();
        fh_tofd_ToF[1]->Draw("colz");
        cTofd_planes->cd(9);
        gPad->SetLogz();
        fh_tofd_multihit[1]->Draw("colz");
        cTofd_planes->cd(10);
        gPad->SetLogz();
        fh_tofd_dt[0]->Draw("colz");

        cTofd_planes->cd(11);
        fh_tofd_channels[2]->Draw();
        cTofd_planes->cd(12);
        gPad->SetLogz();
        fh_tofd_TotPm[2]->Draw("colz");
        cTofd_planes->cd(13);
        gPad->SetLogz();
        fh_tofd_ToF[2]->Draw("colz");
        cTofd_planes->cd(14);
        gPad->SetLogz();
        fh_tofd_multihit[2]->Draw("colz");
        cTofd_planes->cd(15);
        gPad->SetLogz();
        fh_tofd_dt[1]->Draw("colz");

        cTofd_planes->cd(16);
        fh_tofd_channels[3]->Draw();
        cTofd_planes->cd(17);
        gPad->SetLogz();
        fh_tofd_TotPm[3]->Draw("colz");
        cTofd_planes->cd(18);
        gPad->SetLogz();
        fh_tofd_ToF[3]->Draw("colz");
        cTofd_planes->cd(19);
        gPad->SetLogz();
        fh_tofd_multihit[3]->Draw("colz");
        cTofd_planes->cd(20);
        gPad->SetLogz();
        fh_tofd_dt[2]->Draw("colz");

        cTofd_planes->cd(0);
        run->AddObject(cTofd_planes);

        run->GetHttpServer()->RegisterCommand("Reset_TOFD", Form("/Tasks/%s/->Reset_TOFD_Histo()", GetName()));
    }

    // -------------------------------------------------------------------------

    //------------------------------------------------------------------------
    // PSPX detector
    // get access to data

    if (fMappedItems.at(DET_PSPX) || fCalItems.at(DET_PSPX) || fHitItems.at(DET_PSPX))
    {
        TCanvas* cPspx_comp = new TCanvas("Pspx_comp", "Pspx Comparison", 10, 10, 1100, 1000);
        cPspx_comp->Divide(N_PSPX, 3);

        Int_t Emax = 1000000;

        if (fMappedItems.at(DET_PSPX))
        {
            // LOG(INFO) << "Init MappedPspx";

            for (UInt_t i = 0; i < N_PSPX; i++)
            {
                if (i % 2 == 0)
                { // even numbers = read out with energy filter
                    fh_pspx_channel_x[i] =
                        new TH1F(Form("pspx_%d_energyreadout_channel_x", i / 2),
                                 Form("Pspx %d Energy Readout: x Channel;x Position / Channel;Counts", i / 2 + 1),
                                 2 * N_STRIPS_PSPX,
                                 1,
                                 2 * N_STRIPS_PSPX + 1);
                    fh_pspx_channel_y[i] =
                        new TH1F(Form("pspx_%d_energyreadout_channel_y", i / 2),
                                 Form("Pspx %d Energy Readout: y Channel;y Position / Channel;Counts", i / 2 + 1),
                                 2 * N_STRIPS_PSPX,
                                 2 * N_STRIPS_PSPX + 1,
                                 4 * N_STRIPS_PSPX + 1);

                    fh_pspx_multiplicity_x[i] =
                        new TH1F(Form("pspx_%d_energyreadout_multiplicity_x", i / 2),
                                 Form("Pspx %d Energy Readout: x Multiplicity;Multiplicity;Counts", i / 2 + 1),
                                 10,
                                 0,
                                 10);
                    fh_pspx_multiplicity_y[i] =
                        new TH1F(Form("pspx_%d_energyreadout_multiplicity_y", i / 2),
                                 Form("Pspx %d Energy Readout: y Multiplicity;Multiplicity;Counts", i / 2 + 1),
                                 10,
                                 0,
                                 10);
                    fh_pspx_strips_position[i] = new TH2F(
                        Form("pspx_%d_energyreadout_strips", i),
                        Form("Pspx %d Energy Readout Position;x Position / Strips; y Position / Strips", i / 2 + 1),
                        N_STRIPS_PSPX,
                        1,
                        N_STRIPS_PSPX + 1,
                        N_STRIPS_PSPX,
                        N_STRIPS_PSPX + 1,
                        2 * N_STRIPS_PSPX + 1);
                }
                else
                { // odd numbers = read out with position filter
                    fh_pspx_channel_x[i] =
                        new TH1F(Form("pspx_%d_positionreadout_channel_x", i / 2),
                                 Form("Pspx %d Position Readout: x Channel;x Position / Channel;Counts", i / 2 + 1),
                                 2 * N_STRIPS_PSPX,
                                 1,
                                 2 * N_STRIPS_PSPX + 1);
                    fh_pspx_channel_y[i] =
                        new TH1F(Form("pspx_%d_positionreadout_channel_y", i / 2),
                                 Form("Pspx %d Position Readout: y Channel;y Position / Channel;Counts", i / 2 + 1),
                                 2 * N_STRIPS_PSPX,
                                 2 * N_STRIPS_PSPX + 1,
                                 4 * N_STRIPS_PSPX + 1);

                    fh_pspx_multiplicity_x[i] =
                        new TH1F(Form("pspx_%d_positionreadout_multiplicity_x", i / 2),
                                 Form("Pspx %d Position Readout: x Multiplicity;Multiplicity;Counts", i / 2 + 1),
                                 10,
                                 0,
                                 10);
                    fh_pspx_multiplicity_y[i] =
                        new TH1F(Form("pspx_%d_positionreadout_multiplicity_y", i / 2),
                                 Form("Pspx %d Position Readout: y Multiplicity;Multiplicity;Counts", i / 2 + 1),
                                 10,
                                 0,
                                 10);
                    fh_pspx_strips_position[i] = new TH2F(
                        Form("pspx_%d_positionreadout_strips", i),
                        Form("Pspx %d Position Readout Position;x Position / Strips; y Position / Strips", i / 2 + 1),
                        N_STRIPS_PSPX,
                        1,
                        N_STRIPS_PSPX + 1,
                        N_STRIPS_PSPX,
                        N_STRIPS_PSPX + 1,
                        2 * N_STRIPS_PSPX + 1);
                }
            }

            TCanvas* cPspx_strips = new TCanvas("Pspx_strips", "Pspx Strips", 10, 10, 1100, 1000);
            cPspx_strips->Divide(N_PSPX, 2);

            for (UInt_t i = 0; i < N_PSPX; i++)
            {
                cPspx_strips->cd(i + 1);
                fh_pspx_channel_x[i]->Draw();

                cPspx_strips->cd(i + 1 + N_PSPX);
                fh_pspx_channel_y[i]->Draw();
            }

            run->AddObject(cPspx_strips);

            TCanvas* cPspx_multiplicity = new TCanvas("Pspx_multiplicity", "Pspx Multiplicity", 10, 10, 1100, 1000);
            cPspx_multiplicity->Divide(N_PSPX, 2);

            for (UInt_t i = 0; i < N_PSPX; i++)
            {
                cPspx_multiplicity->cd(i + 1);
                fh_pspx_multiplicity_x[i]->Draw();

                cPspx_multiplicity->cd(i + 1 + N_PSPX);
                fh_pspx_multiplicity_y[i]->Draw();
            }

            run->AddObject(cPspx_multiplicity);

            // Fill cPspx_comp with Mapped level data
            for (UInt_t i = 0; i < N_PSPX; i++)
            {
                cPspx_comp->cd(i + 1);
                fh_pspx_strips_position[i]->Draw("colz");

                // Remove the current axis
                fh_pspx_strips_position[i]->GetYaxis()->SetLabelOffset(999);
                fh_pspx_strips_position[i]->GetYaxis()->SetTickLength(0);

                // Redraw the new inverted axis
                gPad->Update();
                TGaxis* newaxis = new TGaxis(gPad->GetUxmin(),
                                             gPad->GetUymax(),
                                             gPad->GetUxmin() - 0.001,
                                             gPad->GetUymin(),
                                             fh_pspx_strips_position[i]->GetYaxis()->GetXmin(),
                                             fh_pspx_strips_position[i]->GetYaxis()->GetXmax(),
                                             510,
                                             "+");
                newaxis->SetLabelOffset(0.003);
                newaxis->SetLabelSize(0.03);
                newaxis->SetTickLength(0.025);
                newaxis->Draw();
            }
        }

        if (fCalItems.at(DET_PSPX))
        {
            UInt_t nbins = 200;

            for (UInt_t i = 0; i < N_PSPX; i++)
            {

                if (i % 2 == 0)
                {
                    fh_pspx_cal_energy_frontback[i] = new TH2F(
                        Form("pspx_%d_energy_frontback", i / 2),
                        Form("Pspx %d Energy Back vs Front;Front Energy / arb.u.;Back Energy / arb.u.", i / 2 + 1),
                        nbins,
                        0,
                        Emax,
                        nbins,
                        -Emax,
                        0);
                }
                else
                {
                    fh_pspx_cal_energy_frontback[i] = new TH2F(
                        Form("pspx_%d_position_frontback", i / 2),
                        Form("Pspx %d Energy Back vs Front;Front Energy / arb.u.;Back Energy / arb.u.", i / 2 + 1),
                        nbins,
                        0,
                        Emax,
                        nbins,
                        -Emax,
                        0);
                }
            }

            // Fill cPspx_comp with Cal level data
            for (UInt_t i = 0; i < N_PSPX; i++)
            {

                cPspx_comp->cd(i + 1 + N_PSPX); // i*2
                fh_pspx_cal_energy_frontback[i]->Draw("colz");
            }
        }

        if (fHitItems.at(DET_PSPX))
        {
            UInt_t nbins = 100;
            Double_t length = 10; // detector length, range of histogram

            for (UInt_t i = 0; i < (N_PSPX + 1) / 2; i++)
            {
                fh_pspx_hit_position[i] = new TH2F(Form("pspx_%d_position_cm", i),
                                                   Form("Pspx %d Position;x Position / cm; y Position / cm", i + 1),
                                                   nbins * 10,
                                                   -length / 2.,
                                                   length / 2.,
                                                   nbins * 10,
                                                   -length / 2.,
                                                   length / 2.);

                fh_pspx_hit_energy[i] = new TH1F(Form("pspx_%d_energy", i),
                                                 Form("Pspx %d Energy;Energy / arb.u.; Counts", i + 1),
                                                 nbins * 10,
                                                 0,
                                                 2 * Emax);

                fh_pspx_hit_multi[i] = new TH2F(
                    Form("pspx_%d_multi", i), Form("Pspx %d Multi;x Multi ; y Multi", i + 1), 20, 0, 20, 20, 0, 20);
            }

            TCanvas* cPspx_hit = new TCanvas("Pspx_hit", "Pspx Hit", 10, 10, 1100, 1000);
            cPspx_hit->Divide((N_PSPX + 1) / 2, 3);

            for (UInt_t i = 0; i < (N_PSPX + 1) / 2; i++)
            {
                cPspx_hit->cd(i + 1);
                gPad->SetLogz();
                fh_pspx_hit_position[i]->Draw("colz");

                cPspx_hit->cd(i + 1 + (N_PSPX + 1) / 2);
                gPad->SetLogy();
                fh_pspx_hit_energy[i]->Draw();

                cPspx_hit->cd(i + 1 + (N_PSPX + 1) / 2 + 3);
                gPad->SetLogz();
                fh_pspx_hit_multi[i]->Draw("colz");
            }

            run->AddObject(cPspx_hit);

            // Fill cPspx_comp with Hit level data-((channel_y[i][0] + 1) / 2) + 3 * N_STRIPS_PSPX + 1)
            for (UInt_t i = 0; i < (N_PSPX + 1) / 2; i++)
            {
                cPspx_comp->cd(i * 2 + 2 * N_PSPX + 1); // supposed to be +2 if
                                                        // energy and position
                                                        // readout is used
                fh_pspx_hit_position[i]->Draw("colz");
            }
        }

        run->AddObject(cPspx_comp);

        run->GetHttpServer()->RegisterCommand("Reset_PSPX", Form("/Tasks/%s/->Reset_PSPX_Histo()", GetName()));

        // -------------------------------------------------------------------------

        return kSUCCESS;
    }
}
void R3BOnlineSpectraLosTest::Reset_TWIN_Histo()
{
    if (fMappedItemsTwim)
    {
        fh1_twim_ESum[0]->Reset();
        fh1_twim_ESum[1]->Reset();
        fh1_twim_ESum[2]->Reset();
        fh2_twim_ESum->Reset();
        fh1_Twimmap_mult->Reset();
    }
}
void R3BOnlineSpectraLosTest::Reset_LOS_Histo()
{

    if (fMappedItems.at(DET_LOS))
    {
        for (Int_t iloscount = 0; iloscount < fNofLosDetectors; iloscount++)
        {
            fh_los_channels[iloscount]->Reset();
            fh_los_tres_MCFD[iloscount]->Reset();
            fh_los_tres_TAMEX[iloscount]->Reset();
            fh_los_pos_ToT[iloscount]->Reset();
            fh_los_tot[iloscount]->Reset();
            fh_los_tot_mean[iloscount]->Reset();
            fh_los_pos_MCFD[iloscount]->Reset();
            fh_los_pos_TAMEX[iloscount]->Reset();
            fh_los_multihit[iloscount]->Reset();
            fh_los_ihit_ToT[iloscount]->Reset();
            fh_losTAMEX_vs_Events[iloscount]->Reset();
            fh_losToT_vs_Events[iloscount]->Reset();
            fh_losMCFD_vs_Events[iloscount]->Reset();
            if (fMappedItemsTwim)
                fh_los_twin[iloscount]->Reset();
            fh_los_rw[iloscount]->Reset();
            fh_los_tot_mean_corr[iloscount]->Reset();
        }
        if (fNofLosDetectors > 1)
        {
            fh_los_ToT_vs_ToT->Reset();
            fh_los_nPart_vs_nPart->Reset();
            fh_dt_los1_los2->Reset();
        }
    }
}

void R3BOnlineSpectraLosTest::Reset_ROLU_Histo()
{
    fh_rolu_channels->Reset();
    fh_rolu_tot->Reset();
}

void R3BOnlineSpectraLosTest::Reset_BMON_Histo()
{
    fhTrigger->Reset();
    fhTpat->Reset();
    fh_spill_length->Reset();
    fh_IC->Reset();
    fh_IC_spill->Reset();
    fh_SEE->Reset();
    fh_SEE_spill->Reset();
    fh_TOFDOR->Reset();
    fh_TOFDOR_spill->Reset();
}

void R3BOnlineSpectraLosTest::Reset_SCI8_Histo()
{
    fh_sci8_channels->Reset();
    fh_sci8_tres_MCFD->Reset();
    fh_sci8_tres_TAMEX->Reset();
    fh_sci8_tot->Reset();
    fh_sci8_tot_mean->Reset();
    fh_sci8_dt_hits->Reset();
    fh_sci8_dt_hits_l->Reset();
    fh_sci8_dt_hits_t->Reset();
    fh_sci8_multihit->Reset();
    fh_sci8_multihitVFTX->Reset();
    fh_sci8_multihitLEAD->Reset();
    fh_sci8_multihitTRAI->Reset();
    fh_tof_sci8->Reset();
}
void R3BOnlineSpectraLosTest::Reset_TOFD_Histo()
{
    for (int i = 0; i < N_PLANE_MAX_TOFD; i++)
    {
        fh_tofd_channels[i]->Reset();
        fh_tofd_multihit[i]->Reset();
        fh_tofd_ToF[i]->Reset();
        fh_tofd_TotPm[i]->Reset();
    }
    fh_tofd_dt[0]->Reset();
    fh_tofd_dt[1]->Reset();
    fh_tofd_dt[2]->Reset();
}
void R3BOnlineSpectraLosTest::Reset_FIBERS_Histo()
{
    if (fMappedItems.at(DET_PSPX))
    {
        for (UInt_t i = 0; i < N_PSPX; i++)
        {
            fh_pspx_channel_x[i]->Reset();
            fh_pspx_channel_y[i]->Reset();
            fh_pspx_multiplicity_x[i]->Reset();
            fh_pspx_multiplicity_y[i]->Reset();
            fh_pspx_strips_position[i]->Reset();
        }
    }
    if (fCalItems.at(DET_PSPX))
    {
        for (UInt_t i = 0; i < N_PSPX; i++)
        {
            fh_pspx_cal_energy_frontback[i]->Reset();
        }
    }
    if (fHitItems.at(DET_PSPX))
    {
        for (UInt_t i = 0; i < N_PSPX / 2; i++)
        {
            fh_pspx_hit_position[i]->Reset();
            fh_pspx_hit_energy[i]->Reset();
        }
    }
    for (Int_t ifibcount = 0; ifibcount < NOF_FIB_DET; ifibcount++)
    {
        if (fMappedItems.at(DET_FI_FIRST + ifibcount))
        {
            fh_channels_Fib[ifibcount]->Reset();
            fh_multihit_m_Fib[ifibcount]->Reset();
            fh_multihit_s_Fib[ifibcount]->Reset();
            fh_fibers_Fib[ifibcount]->Reset();
            fh_mult_Fib[ifibcount]->Reset();
            fh_ToT_m_Fib[ifibcount]->Reset();
            fh_ToT_s_Fib[ifibcount]->Reset();
            fh_ToT_single_Fib[ifibcount]->Reset();
            fh_Fib_ToF[ifibcount]->Reset();
            fh_Fib_pos[ifibcount]->Reset();
            fh_Fib_vs_Events[ifibcount]->Reset();
            fh_channels_single_Fib[ifibcount]->Reset();
        }
    }
}

void R3BOnlineSpectraLosTest::Exec(Option_t* option)
{
    //  cout << "fNEvents " << fNEvents << endl;

    FairRootManager* mgr = FairRootManager::Instance();
    if (NULL == mgr)
    {
        // FairLogger::GetLogger()->Fatal(MESSAGE_ORIGIN, "FairRootManager not found");
        LOG(ERROR) << "FairRootManager not found";
        return;
    }

    ofstream myFile("data_eng_run40.dat", ios_base::out | ios_base::app);

    time = header->GetTimeStamp() / 1.6; // divided by 1.6 for stand alone daq with Vulom generating time stamps

    // time = 0;

    if (time_start == 0 && time > 0)
    {
        time_start = time;
        fNEvents_start = fNEvents;
    }

    if (header->GetTrigger() == 12)
        time_spill_start = time; // header->GetTimeStamp();    // spill start in nsec
    if (header->GetTrigger() == 13)
        time_spill_end = time; // header->GetTimeStamp();    // spill end  in nsec

    if (header->GetTrigger() == 12)
        cout << "Spill start: " << double(time_spill_start - time_start) / 1.e9 << " sec" << endl;
    if (header->GetTrigger() == 13)
        cout << "Spill stop: " << double(time_spill_end - time_start) / 1.e9 << " sec" << endl;

    fhTrigger->Fill(header->GetTrigger());

    if ((fTrigger >= 0) && (header) && (header->GetTrigger() != fTrigger))
        return;

    Int_t tpatbin;
    for (int i = 0; i < 16; i++)
    {
        tpatbin = (header->GetTpat() & (1 << i));
        if (tpatbin != 0)
            fhTpat->Fill(i + 1);
    }

    // fTpat = 1-16; fTpat_bit = 0-15
    Int_t fTpat_bit = fTpat - 1;
    Int_t itpat;
    Int_t tpatvalue;
    if (fTpat_bit >= 0)
    {
        itpat = header->GetTpat();
        tpatvalue = (itpat && (1 << fTpat_bit)) >> fTpat_bit;
        if ((tpatvalue == 0))
            return;
    }

    if (fMappedItems.at(DET_SAMPLER))
    {

        auto det = fMappedItems.at(DET_SAMPLER);
        Int_t nHitsSamp = det->GetEntriesFast();
        for (Int_t ihit = 0; ihit < nHitsSamp; ihit++)
        {
            auto hit = (R3BSamplerMappedData*)det->At(ihit);
            // time is in steps of 10 ns
            // is is a 34 bit number, so max 1073741823
            samplerCurr = hit->GetTime();
            //		cout<<"Sampler vs Tstemp: " << samplerCurr<<", "<<time<<", "<<fNEvents<<endl;

            /*
                        if(spill_on_sampler && ihit == 0) {
                            samplerSpill = samplerCurr;
                            spill_on_sampler=false;
                        }


                        long samp = (samplerCurr - samplerSpill); //time in 10 ns
                        if(samp<0) samp += 1073741823;
                        if((double)samp/1e8 > fSpillLength && spillCounter>0) {
                            cout <<"Missed spill end!" << endl;
                            std::cout << samplerCurr << "  " << samplerPrev << "  " << samplerSpill<< "  " <<
               (double)samp/1e8 << '\n';
                        }
            */
            Double_t dt = ((double)(samplerCurr - samplerPrev)) / 100.;

            samplerPrev = samplerCurr;
        }
    }

    if (fMappedItems.at(DET_BMON))
    {

        Bool_t spectra_clear = false;
        Double_t xtime = double(time - time_start) / 1.e9;
        // for reseting spectra
        Int_t icount = iclear_count * reset_time; // reset after reset_time (sec)
        if (time_clear < 0. && int(xtime) % icount == 0 && xtime > 1.)
        {
            time_clear = xtime;
            spectra_clear = true;
        }
        unsigned long IC;
        unsigned long SEETRAM;
        unsigned long TOFDOR;
        Bool_t bmon_read = false;

        auto det = fMappedItems.at(DET_BMON);
        Int_t nHitsbm = det->GetEntriesFast();

        for (Int_t ihit = 0; ihit < nHitsbm; ihit++)
        {
            R3BBeamMonitorMappedData* hit = (R3BBeamMonitorMappedData*)det->At(ihit);
            if (!hit)
                continue;

            IC = hit->GetIC();           // negative values if offset not high enough
            SEETRAM = hit->GetSEETRAM(); // negative values if offset not high enough
            TOFDOR = hit->GetTOFDOR();   // only positive values possible

            //     unsigned long long time = header->GetTimeStamp();

            if (fNEvents == fNEvents_start)
            {
                see_mem = SEETRAM;
                ic_mem = IC;
                tofdor_mem = TOFDOR;
                time_mem = time_start;
                see_start = SEETRAM;
                ic_start = IC;
                tofdor_start = TOFDOR;
                time_prev_read = time_start;
            }

            if (time > 0)
            {

                fh_spill_length->Fill((time - time_mem) / 1e9);

                // Spectra below are filled every read_time (secs)
                if (time_to_read == 0 && (time - time_prev_read) >= read_time * 1000000000)
                {
                    time_to_read = time;
                    bmon_read = true;
                }

                if (bmon_read)
                {

                    tdiff = double(time - time_mem) / 1.e9;
                    fNorm = 1.e-3 / (double(time - time_prev_read) / 1.e9); // kHz

                    // IC:
                    Int_t yIC = IC - ic_start;
                    fh_IC->Fill(tdiff, yIC);
                    fh_IC_spill->Fill(tdiff, (IC - ic_mem) * fNorm);
                    ic_mem = IC;

                    // SEETRAM:
                    Int_t ySEE = SEETRAM - see_start;
                    fh_SEE->Fill(tdiff, ySEE);

                    Double_t ySEE_part = (SEETRAM - see_mem) * fNorm * 1.e+3 - see_offset * calib_SEE;
                    fh_SEE_spill->Fill(tdiff, ySEE_part);
                    see_mem = SEETRAM;

                    // TOFDOR:
                    Int_t yTOFDOR = TOFDOR - tofdor_start;
                    fh_TOFDOR->Fill(tdiff, yTOFDOR);
                    fh_TOFDOR_spill->Fill(tdiff, (TOFDOR - tofdor_mem) * fNorm);
                    tofdor_mem = TOFDOR;

                    time_to_read = 0;
                    time_prev_read = time;
                    bmon_read = false;
                }

                if (spectra_clear)
                {
                    fh_spill_length->Reset();
                    fh_IC_spill->Reset("ICESM");
                    fh_SEE_spill->Reset("ICESM");
                    fh_TOFDOR_spill->Reset("ICESM");
                    fh_IC->Reset("ICESM");
                    fh_SEE->Reset("ICESM");
                    fh_TOFDOR->Reset("ICESM");
                    time_mem = time;
                    time_clear = -1.;
                    iclear_count = iclear_count + 1;
                    spectra_clear = false;
                    see_start = SEETRAM;
                    ic_start = IC;
                    tofdor_start = TOFDOR;
                }
            }
        }
    }

    // **************   ROLU *************************************

    if (fMappedItems.at(DET_ROLU))
    {
        auto det = fMappedItems.at(DET_ROLU);
        Int_t nHits = det->GetEntriesFast();

        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BRoluMappedData* hit = (R3BRoluMappedData*)det->At(ihit);
            if (!hit)
                continue;

            // channel numbers are stored 1-based (1..n)
            Int_t iDet = hit->GetDetector(); // 1..
            Int_t iCha = hit->GetChannel();  // 1..

            if (iDet < 2)
                fh_rolu_channels->Fill(iCha); // ROLU 1
            if (iDet > 1)
                fh_rolu_channels->Fill(iCha + 4); // ROLU 2
        }
    }

    Int_t nParts;
    if (fCalItems.at(DET_ROLU))
    {

        auto det = fCalItems.at(DET_ROLU);
        nParts = det->GetEntriesFast();

        if (nParts < 1)
            return;

        Int_t iDet = 0;
        Double_t timeRolu_L[nParts][2][4] = { 0.0 / 0.0 };
        Double_t timeRolu_T[nParts][2][4] = { 0.0 / 0.0 };
        Double_t totRolu[nParts][2][4] = { 0.0 / 0.0 };

        for (Int_t iPart = 0; iPart < nParts; iPart++)
        {
            /*
             * nParts is the number of particle passing through detector in one event
             */
            R3BRoluCalData* calData = (R3BRoluCalData*)det->At(iPart);
            iDet = calData->GetDetector();

            for (Int_t iCha = 0; iCha < 4; iCha++)
            {

                if (!(IS_NAN(calData->GetTimeL_ns(iCha))))
                { // TAMEX leading
                    timeRolu_L[iPart][iDet - 1][iCha] = calData->GetTimeL_ns(iCha);
                }
                if (!(IS_NAN(calData->GetTimeT_ns(iCha))))
                { // TAMEX trailing
                    timeRolu_T[iPart][iDet - 1][iCha] = calData->GetTimeT_ns(iCha);
                }

                //  cout<<"ROLU: "<<iPart<<", "<<iCha<<"; "<<timeRolu_L[iPart][iCha]<<",
                //  "<<timeRolu_T[iPart][iCha]<<endl;

                if (timeRolu_T[iPart][iDet - 1][iCha] > 0. && timeRolu_L[iPart][iDet - 1][iCha] > 0. &&
                    !(IS_NAN(timeRolu_T[iPart][iDet - 1][iCha])) && !(IS_NAN(timeRolu_L[iPart][iDet - 1][iCha])))
                {
                    while (timeRolu_T[iPart][iDet - 1][iCha] - timeRolu_L[iPart][iDet - 1][iCha] <= 0.)
                    {
                        timeRolu_T[iPart][iDet - 1][iCha] = timeRolu_T[iPart][iDet - 1][iCha] + 2048. * fClockFreq;
                    }

                    totRolu[iPart][iDet - 1][iCha] =
                        timeRolu_T[iPart][iDet - 1][iCha] - timeRolu_L[iPart][iDet - 1][iCha];
                }

                if (iDet < 2)
                    fh_rolu_tot->Fill(iCha + 1, totRolu[iPart][iDet - 1][iCha]);
                if (iDet > 1)
                    fh_rolu_tot->Fill(iCha + 5, totRolu[iPart][iDet - 1][iCha]);
            }

            if (!calData)
            {
                cout << "Rolu !calData" << endl;
                continue; // can this happen?
            }
        }
    }
    // TWIN MUSIC
    Double_t dE_TWIN;
    Bool_t twin = false;
    // Fill mapped data
    if (fMappedItemsTwim && fMappedItemsTwim->GetEntriesFast() > 0)
    {
        Double_t e1 = 0., e2 = 0., E[NbSections][NbAnodes], DT[NbSections][NbAnodes + NbSections];
        Double_t n1 = 0., n2 = 0.;
        for (Int_t j = 0; j < NbSections; j++)
            for (Int_t i = 0; i < NbAnodes; i++)
                E[j][i] = 0.; // mult=1 !!!
        for (Int_t j = 0; j < NbSections; j++)
            for (Int_t i = 0; i < NbAnodes + NbSections; i++)
                DT[j][i] = 0.; // mult=1 !!!
        Int_t nHits = fMappedItemsTwim->GetEntriesFast();
        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BSofTwimMappedData* hit = (R3BSofTwimMappedData*)fMappedItemsTwim->At(ihit);
            if (!hit)
                continue;
            fh1_Twimmap_mult->Fill(hit->GetAnodeID());
            //  fh1_twimmap_E[hit->GetSecID()][hit->GetAnodeID()]->Fill(hit->GetEnergy());
            if (DT[hit->GetSecID()][hit->GetAnodeID()] == 0 && E[hit->GetSecID()][hit->GetAnodeID()] == 0 &&
                hit->GetEnergy() < 8192 && hit->GetEnergy() > 0)
            {
                E[hit->GetSecID()][hit->GetAnodeID()] = hit->GetEnergy(); // mult=1 !!!
                DT[hit->GetSecID()][hit->GetAnodeID()] = hit->GetTime();  // mult=1 !!!
            }

            if (hit->GetAnodeID() < NbAnodes / 2 && hit->GetAnodeID() != 7)
            {
                e1 = e1 + hit->GetEnergy();
                n1++;
            }
            else if (hit->GetAnodeID() >= NbAnodes / 2 && hit->GetAnodeID() < NbAnodes && hit->GetAnodeID() != 15)
            {
                e2 = e2 + hit->GetEnergy();
                n2++;
            }
        }
        if ((e1 / n1 < 5000) && abs(e1 / n1 - e2 / n2) < 200.)
        {
            twin = true;
            fh1_twim_ESum[0]->Fill(e1 / n1);
            fh1_twim_ESum[1]->Fill(e2 / n2);
            fh1_twim_ESum[2]->Fill((e1 + e2) / (n1 + n2));
            fh2_twim_ESum->Fill(e1 / n1, e2 / n2);
        }
        // cout<<"TWIN: "<< e1/n1<<", "<<e2/n2<<endl;
        dE_TWIN = (e1 + e2) / (n1 + n2);
    }

    //----------------------------------------------------------------------
    // LOS detector
    //----------------------------------------------------------------------
    // Time nomenclature for LOS:
    // lt=0, l=1,lb=2,b=3,rb=4,r=5,rt=6,t=7
    //   t / \lt
    //   rt| |l
    //   r | |lb
    //   rb\ /b

    Double_t timeTofd = 0;

    Double_t timeLosV[fNofLosDetectors][32];
    Double_t LosTresV[fNofLosDetectors][32];
    Double_t timeLosT[fNofLosDetectors][32];
    Double_t LosTresT[fNofLosDetectors][32];
    Double_t timeLos[fNofLosDetectors][32];
    Double_t totsum[fNofLosDetectors][32];
    Double_t xT_cm[fNofLosDetectors][32];
    Double_t yT_cm[fNofLosDetectors][32];
    Double_t xToT_cm[fNofLosDetectors][32];
    Double_t yToT_cm[fNofLosDetectors][32];
    Double_t xV_cm[fNofLosDetectors][32];
    Double_t yV_cm[fNofLosDetectors][32];

    Double_t time_V[fNofLosDetectors][32][8]; // [det][multihit][pm]
    Double_t time_L[fNofLosDetectors][32][8];
    Double_t time_T[fNofLosDetectors][32][8];
    Double_t tot[fNofLosDetectors][32][8];
    Double_t time_MTDC[32][8] = { 0. };
    Double_t LosTresMTDC[32];

    for (Int_t idet = 0; idet < fNofLosDetectors; idet++)
    {
        for (Int_t imult = 0; imult < 32; imult++)
        {
            timeLosV[idet][imult] = 0.0;
            LosTresV[idet][imult] = 0.0 / 0.0;
            timeLosT[idet][imult] = 0.0;
            LosTresT[idet][imult] = 0.0 / 0.0;
            timeLos[idet][imult] = 0.0;
            totsum[idet][imult] = 0.0;
            xT_cm[idet][imult] = 0.0 / 0.0;
            yT_cm[idet][imult] = 0.0 / 0.0;
            xToT_cm[idet][imult] = 0.0 / 0.0;
            yToT_cm[idet][imult] = 0.0 / 0.0;
            xV_cm[idet][imult] = 0.0 / 0.0;
            yV_cm[idet][imult] = 0.0 / 0.0;
            for (Int_t icha = 0; icha < 8; icha++)
            {
                time_V[idet][imult][icha] = 0.0 / 0.0; // [det][multihit][pm]
                time_L[idet][imult][icha] = 0.0 / 0.0;
                time_T[idet][imult][icha] = 0.0 / 0.0;
                tot[idet][imult][icha] = 0.0 / 0.0;
            }
        }
    }
    Int_t nPartLOS = 0;
    Int_t nPartc[fNofLosDetectors];
    for (Int_t d = 0; d < fNofLosDetectors; d++)
        nPartc[d] = 0;
    Bool_t iLOSType[fNofLosDetectors][32];
    for (Int_t idet = 0; idet < fNofLosDetectors; idet++)
    {
        for (Int_t imult = 0; imult < 33; imult++)
        {
            iLOSType[idet][imult] = false;
        }
    }

    Int_t Multip;

    if (fMappedItems.at(DET_LOS))
    {
        auto det = fMappedItems.at(DET_LOS);
        Int_t nHits = det->GetEntriesFast();

        Multip = nHits;

        //    cout<<"LOS: nHits"<<nHits<<endl;

        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BLosMappedData* hit = (R3BLosMappedData*)det->At(ihit);
            if (!hit)
                continue;

            // channel numbers are stored 1-based (1..n)
            Int_t iDet = hit->GetDetector(); // 1..
            Int_t iCha = hit->GetChannel();  // 1..
            Int_t iTyp = hit->GetType();     // 0,1,2,3

            if (iTyp < 3)
                fh_los_channels[iDet - 1]->Fill(iCha); // exclude MTDC data
        }
    }

    if (fCalItems.at(DET_LOS))
    {
        auto det = fCalItems.at(DET_LOS);
        nPartLOS = det->GetEntriesFast();

        Int_t iDet = 0;
        Double_t time_V_LOS1[32][8] = { 0. };
        Double_t time_V_LOS2[32][8] = { 0. };

        for (Int_t iPart = 0; iPart < nPartLOS; iPart++)
        {
            /*
             * nPart is the number of particle passing through LOS detector in one event
             */
            R3BLosCalData* calData = (R3BLosCalData*)det->At(iPart);
            iDet = calData->GetDetector();

            Double_t sumvtemp = 0, sumltemp = 0, sumttemp = 0;
            for (Int_t iCha = 0; iCha < 8; iCha++)
            {
                sumvtemp += calData->GetTimeV_ns(iCha);
                sumltemp += calData->GetTimeL_ns(iCha);
                sumttemp += calData->GetTimeT_ns(iCha);
            }
            if (!(IS_NAN(sumvtemp)) && !(IS_NAN(sumltemp)) && !(IS_NAN(sumltemp)))
            {
                nPartc[iDet - 1]++;
            }
            else
            {
                continue;
            }

            for (Int_t iCha = 0; iCha < 8; iCha++)
            {
                if (iDet == 1)
                {
                    time_V_LOS1[nPartc[iDet - 1] - 1][iCha] = 0. / 0.;
                    if (!(IS_NAN(calData->GetTimeV_ns(iCha))))
                    { // VFTX
                        time_V_LOS1[nPartc[iDet - 1] - 1][iCha] = calData->GetTimeV_ns(iCha);
                    }
                    if (!(IS_NAN(calData->GetTimeL_ns(iCha))))
                    { // TAMEX leading
                        time_L[iDet - 1][nPartc[iDet - 1] - 1][iCha] = calData->GetTimeL_ns(iCha);
                    }

                    if (!(IS_NAN(calData->GetTimeT_ns(iCha))))
                    { // TAMEX trailing
                        time_T[iDet - 1][nPartc[iDet - 1] - 1][iCha] = calData->GetTimeT_ns(iCha);
                    }
                    time_MTDC[nPartc[iDet - 1]][iCha] = 0. / 0.;
                    if (!(IS_NAN(calData->GetTimeM_ns(iCha))))
                    { // MTDC
                        time_MTDC[nPartc[iDet - 1] - 1][iCha] = calData->GetTimeM_ns(iCha);
                    }
                }
                if (iDet == 2)
                {
                    time_V_LOS2[nPartc[iDet - 1] - 1][iCha] = 0. / 0.;
                    if (!(IS_NAN(calData->GetTimeV_ns(iCha))))
                    { // VFTX
                        time_V_LOS2[nPartc[iDet - 1] - 1][iCha] = calData->GetTimeV_ns(iCha);
                    }
                    if (!(IS_NAN(calData->GetTimeL_ns(iCha))))
                    { // TAMEX leading
                        time_L[iDet - 1][nPartc[iDet - 1] - 1][iCha] = calData->GetTimeL_ns(iCha);
                    }
                    if (!(IS_NAN(calData->GetTimeT_ns(iCha))))
                    { // TAMEX trailing
                        time_T[iDet - 1][nPartc[iDet - 1] - 1][iCha] = calData->GetTimeT_ns(iCha);
                    }
                }
            }

            if (!calData)
            {
                cout << "LOS !calData" << endl;
                continue; // can this happen?
            }
        }

        // Sorting VFTX data:

        // detector 1
        if (nPartc[0] > 0)
        {
            std::qsort(time_V_LOS1, nPartc[0], sizeof(*time_V_LOS1), [](const void* arg1, const void* arg2) -> int {
                double const* lhs = static_cast<double const*>(arg1);
                double const* rhs = static_cast<double const*>(arg2);

                return (lhs[0] < rhs[0]) ? -1 : ((rhs[0] < lhs[0]) ? 1 : 0);
            });
            for (Int_t iPart = 0; iPart < nPartc[0]; iPart++)
            {
                for (int ipm = 0; ipm < 8; ipm++)
                {
                    time_V[0][iPart][ipm] = time_V_LOS1[iPart][ipm];
                }
            }
        }

        // detector 2
        if (fNofLosDetectors > 1 && nPartc[1] > 0)
        {
            std::qsort(time_V_LOS2, nPartc[1], sizeof(*time_V_LOS2), [](const void* arg1, const void* arg2) -> int {
                double const* lhs = static_cast<double const*>(arg1);
                double const* rhs = static_cast<double const*>(arg2);

                return (lhs[0] < rhs[0]) ? -1 : ((rhs[0] < lhs[0]) ? 1 : 0);
            });
            for (Int_t iPart = 0; iPart < nPartc[1]; iPart++)
            {
                for (int ipm = 0; ipm < 8; ipm++)
                {
                    time_V[1][iPart][ipm] = time_V_LOS2[iPart][ipm];
                }
            }
        }

        // End sorting

        std::vector<double_t> time_first(fNofLosDetectors, -1.);
        std::vector<double_t> time0(fNofLosDetectors, -1.);
        std::vector<double_t> time1(fNofLosDetectors, -1.);
        std::vector<double_t> time_abs(fNofLosDetectors, -1.);

        for (iDet = 1; iDet <= fNofLosDetectors; iDet++)
        {
            for (Int_t iPart = 0; iPart < nPartc[iDet - 1]; iPart++)
            {
                Bool_t iLOSTypeMCFD = false;
                Bool_t iLOSTypeTAMEX = false;

                if (time_V[iDet - 1][iPart][0] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][0])) &&
                    time_V[iDet - 1][iPart][1] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][1])) &&
                    time_V[iDet - 1][iPart][2] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][2])) &&
                    time_V[iDet - 1][iPart][3] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][3])) &&
                    time_V[iDet - 1][iPart][4] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][4])) &&
                    time_V[iDet - 1][iPart][5] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][5])) &&
                    time_V[iDet - 1][iPart][6] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][6])) &&
                    time_V[iDet - 1][iPart][7] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][7])))
                {
                    iLOSTypeMCFD = true; // all 8 MCFD times
                }

                if (time_L[iDet - 1][iPart][0] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][0])) &&
                    time_L[iDet - 1][iPart][1] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][1])) &&
                    time_L[iDet - 1][iPart][2] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][2])) &&
                    time_L[iDet - 1][iPart][3] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][3])) &&
                    time_L[iDet - 1][iPart][4] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][4])) &&
                    time_L[iDet - 1][iPart][5] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][5])) &&
                    time_L[iDet - 1][iPart][6] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][6])) &&
                    time_L[iDet - 1][iPart][7] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][7])) &&

                    time_T[iDet - 1][iPart][0] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][0])) &&
                    time_T[iDet - 1][iPart][1] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][1])) &&
                    time_T[iDet - 1][iPart][2] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][2])) &&
                    time_T[iDet - 1][iPart][3] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][3])) &&
                    time_T[iDet - 1][iPart][4] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][4])) &&
                    time_T[iDet - 1][iPart][5] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][5])) &&
                    time_T[iDet - 1][iPart][6] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][6])) &&
                    time_T[iDet - 1][iPart][7] > 0. && !(IS_NAN(time_T[iDet - 1][iPart][7])))
                {
                    iLOSTypeTAMEX = true; // all 8 leading and trailing times
                }

                // We will consider only events in which booth MCFD and TAMEX see same number of channels:
                if (iLOSTypeTAMEX && iLOSTypeMCFD)
                    iLOSType[iDet - 1][iPart] = true;

                // if(iDet < fNofLosDetectors+1)
                // {

                if (iLOSType[iDet - 1][iPart])
                {
                    int nPMT = 0;
                    int nPMV = 0;

                    for (int ipm = 0; ipm < 8; ipm++)
                    {

                        if (time_T[iDet - 1][iPart][ipm] > 0. && time_L[iDet - 1][iPart][ipm] > 0. &&
                            !(IS_NAN(time_T[iDet - 1][iPart][ipm])) && !(IS_NAN(time_L[iDet - 1][iPart][ipm])))
                        {
                            while (time_T[iDet - 1][iPart][ipm] - time_L[iDet - 1][iPart][ipm] <= 0.)
                            {
                                time_T[iDet - 1][iPart][ipm] = time_T[iDet - 1][iPart][ipm] + 2048. * fClockFreq;
                            }

                            nPMT = nPMT + 1;
                            tot[iDet - 1][iPart][ipm] = time_T[iDet - 1][iPart][ipm] - time_L[iDet - 1][iPart][ipm];

                            if (tot[iDet - 1][iPart][ipm] > 100.)
                                iLOSType[iDet - 1][iPart] = false; // pileup rejection
                        }

                        if (tot[iDet - 1][iPart][ipm] != 0. && !(IS_NAN(tot[iDet - 1][iPart][ipm])))
                            totsum[iDet - 1][iPart] += tot[iDet - 1][iPart][ipm];

                        if (time_L[iDet - 1][iPart][ipm] > 0. && !(IS_NAN(time_L[iDet - 1][iPart][ipm])))
                            timeLosT[iDet - 1][iPart] += time_L[iDet - 1][iPart][ipm];

                        // Calculate detector time
                        if (time_V[iDet - 1][iPart][ipm] > 0. && !(IS_NAN(time_V[iDet - 1][iPart][ipm])))
                        {
                            timeLosV[iDet - 1][iPart] += time_V[iDet - 1][iPart][ipm];
                            nPMV = nPMV + 1;
                        }

                        if (fNEvents < 5000000)
                            myFile << setprecision(10) << iDet << " " << iPart << " " << ipm << " "
                                   << time_V[iDet - 1][iPart][ipm] << " " << time_L[iDet - 1][iPart][ipm] << " "
                                   << tot[iDet - 1][iPart][ipm] << endl;
                    }

                    totsum[iDet - 1][iPart] = totsum[iDet - 1][iPart] / nPMT;

                    timeLosV[iDet - 1][iPart] = timeLosV[iDet - 1][iPart] / nPMV;

                    timeLosT[iDet - 1][iPart] = timeLosT[iDet - 1][iPart] / nPMT;

                    timeLos[iDet - 1][iPart] = timeLosV[iDet - 1][iPart];

                    // cout<<"LOS: "<<iDet<<", "<<iPart<<"; "<<timeLos[iDet-1][iPart]<<endl;

                    LosTresV[iDet - 1][iPart] = ((time_V[iDet - 1][iPart][0] + time_V[iDet - 1][iPart][2] +
                                                  time_V[iDet - 1][iPart][4] + time_V[iDet - 1][iPart][6]) -
                                                 (time_V[iDet - 1][iPart][1] + time_V[iDet - 1][iPart][3] +
                                                  time_V[iDet - 1][iPart][5] + time_V[iDet - 1][iPart][7])) /
                                                4.;

                    LosTresT[iDet - 1][iPart] = ((time_L[iDet - 1][iPart][0] + time_L[iDet - 1][iPart][2] +
                                                  time_L[iDet - 1][iPart][4] + time_L[iDet - 1][iPart][6]) -
                                                 (time_L[iDet - 1][iPart][1] + time_L[iDet - 1][iPart][3] +
                                                  time_L[iDet - 1][iPart][5] + time_L[iDet - 1][iPart][7])) /
                                                4.;

                    //   LosTresMTDC[iPart] =
                    //   ((time_MTDC[iPart][0]+time_MTDC[iPart][2]+time_MTDC[iPart][4]+time_MTDC[iPart][6]) -
                    //      (time_MTDC[iPart][1]+time_MTDC[iPart][3]+time_MTDC[iPart][5]+time_MTDC[iPart][7]))/4.;

                    LosTresMTDC[iPart] =
                        ((time_MTDC[iPart][1] + time_MTDC[iPart][3]) - (time_MTDC[iPart][5] + time_MTDC[iPart][7])) /
                        4.;

                    // Position from tamex:
                    xT_cm[iDet - 1][iPart] = (time_L[iDet - 1][iPart][1] + time_L[iDet - 1][iPart][2]) / 2. -
                                             (time_L[iDet - 1][iPart][5] + time_L[iDet - 1][iPart][6]) / 2.;
                    yT_cm[iDet - 1][iPart] = (time_L[iDet - 1][iPart][3] + time_L[iDet - 1][iPart][4]) / 2. -
                                             (time_L[iDet - 1][iPart][7] + time_L[iDet - 1][iPart][0]) / 2.;
                    xT_cm[iDet - 1][iPart] = (xT_cm[iDet - 1][iPart] - flosOffsetXT[iDet - 1]) * flosVeffXT[iDet - 1];
                    yT_cm[iDet - 1][iPart] = (yT_cm[iDet - 1][iPart] - flosOffsetYT[iDet - 1]) * flosVeffYT[iDet - 1];

                    // Position from VFTX:
                    xV_cm[iDet - 1][iPart] = (time_V[iDet - 1][iPart][1] + time_V[iDet - 1][iPart][2]) / 2. -
                                             (time_V[iDet - 1][iPart][5] + time_V[iDet - 1][iPart][6]) / 2.;
                    yV_cm[iDet - 1][iPart] = (time_V[iDet - 1][iPart][3] + time_V[iDet - 1][iPart][4]) / 2. -
                                             (time_V[iDet - 1][iPart][7] + time_V[iDet - 1][iPart][0]) / 2.;
                    xV_cm[iDet - 1][iPart] = (xV_cm[iDet - 1][iPart] - flosOffsetXV[iDet - 1]) * flosVeffXV[iDet - 1];
                    yV_cm[iDet - 1][iPart] = (yV_cm[iDet - 1][iPart] - flosOffsetYV[iDet - 1]) * flosVeffYV[iDet - 1];

                    // Position from ToT:
                    xToT_cm[iDet - 1][iPart] = (((tot[iDet - 1][iPart][5] + tot[iDet - 1][iPart][6]) / 2. -
                                                 (tot[iDet - 1][iPart][1] + tot[iDet - 1][iPart][2]) / 2.) /
                                                ((tot[iDet - 1][iPart][1] + tot[iDet - 1][iPart][2] +
                                                  tot[iDet - 1][iPart][5] + tot[iDet - 1][iPart][6]) /
                                                 4.));

                    yToT_cm[iDet - 1][iPart] = (((tot[iDet - 1][iPart][0] + tot[iDet - 1][iPart][7]) / 2. -
                                                 (tot[iDet - 1][iPart][3] + tot[iDet - 1][iPart][4]) / 2.) /
                                                ((tot[iDet - 1][iPart][7] + tot[iDet - 1][iPart][0] +
                                                  tot[iDet - 1][iPart][3] + tot[iDet - 1][iPart][4]) /
                                                 4.));

                    xToT_cm[iDet - 1][iPart] =
                        (xToT_cm[iDet - 1][iPart] - flosOffsetXQ[iDet - 1]) * flosVeffXQ[iDet - 1];
                    yToT_cm[iDet - 1][iPart] =
                        (yToT_cm[iDet - 1][iPart] - flosOffsetYQ[iDet - 1]) * flosVeffYQ[iDet - 1];

                    // time difference between two hits
                    if (time0[iDet - 1] < 0)
                        time0[iDet - 1] = timeLosV[iDet - 1][iPart];
                    time1[iDet - 1] = timeLosV[iDet - 1][iPart];
                    time_abs[iDet - 1] = time1[iDet - 1] - time0[iDet - 1] + (double)(time - time_prev[iDet - 1]);
                    if (time_abs[iDet - 1] > 0 && time_abs[iDet - 1] < 5.E8)
                    {
                        fh_los_dt_hits_ToT[iDet - 1]->Fill(time_abs[iDet - 1] / 1000., totsum[iDet - 1][iPart]);
                    }

                    time_prev[iDet - 1] = time;
                    time0[iDet - 1] = time1[iDet - 1];

                    if (timeLosV[iDet - 1][iPart] > 0. && timeLosV[iDet - 1][iPart] < 8192. * 5. &&
                        !(IS_NAN(timeLosV[iDet - 1][iPart])))
                    {
                        while (timeLosT[iDet - 1][iPart] - timeLosV[iDet - 1][iPart] < 2048. * 5. / 2.)
                        {
                            timeLosT[iDet - 1][iPart] = timeLosT[iDet - 1][iPart] + 2048. * 5.;
                        }
                        while (timeLosT[iDet - 1][iPart] - timeLosV[iDet - 1][iPart] > 2048. * 5. / 2.)
                        {
                            timeLosT[iDet - 1][iPart] = timeLosT[iDet - 1][iPart] - 2048. * 5.;
                        }

                        if (twin)
                            fh_los_vftx_tamex[iDet - 1]->Fill(timeLosT[iDet - 1][iPart] - timeLosV[iDet - 1][iPart]);
                    }

                    fh_losMCFD_vs_Events[iDet - 1]->Fill(fNEvents, LosTresV[iDet - 1][iPart]);
                    fh_losTAMEX_vs_Events[iDet - 1]->Fill(fNEvents, LosTresT[iDet - 1][iPart]);
                    fh_losToT_vs_Events[iDet - 1]->Fill(fNEvents, totsum[iDet - 1][iPart]);

                    if (foptcond == 0)
                    {
                        fh_los_tot_mean[iDet - 1]->Fill(totsum[iDet - 1][iPart]);
                        for (int ipm = 0; ipm < 8; ipm++)
                        {
                            fh_los_tot[iDet - 1]->Fill(ipm + 1, tot[iDet - 1][iPart][ipm]);
                        }

                        fh_los_tres_MCFD[iDet - 1]->Fill(LosTresV[iDet - 1][iPart]);
                        fh_los_tres_TAMEX[iDet - 1]->Fill(LosTresT[iDet - 1][iPart]);
                        //   if(iDet==1) fh_los_tres_MTDC->Fill(LosTresMTDC[iPart]);

                        fh_los_pos_MCFD[iDet - 1]->Fill(xV_cm[iDet - 1][iPart], yV_cm[iDet - 1][iPart]);
                        fh_los_pos_TAMEX[iDet - 1]->Fill(xT_cm[iDet - 1][iPart], yT_cm[iDet - 1][iPart]);
                        fh_los_pos_ToT[iDet - 1]->Fill(xToT_cm[iDet - 1][iPart], yToT_cm[iDet - 1][iPart]);
                    }

                    //  fh_los_ihit_ToT[iDet-1]->Fill(iPart+1,totsum[iDet-1][iPart]);
                    fh_los_multihit[iDet - 1]->Fill(iPart + 1);

                } // if iLosType

                // }
                // else
                // {
                // cout<<"Wrong detector ID for LOS!"<<endl;
                // }
            } // for iPart
        }     // for iDet

    } // if fCallItems

    if (fCalItems.at(DET_LOS) && fNofLosDetectors == 2)
    {
        fh_los_nPart_vs_nPart->Fill(nPartc[0], nPartc[1]);

        for (Int_t iPart1 = 0; iPart1 < nPartc[0]; iPart1++)
        {
            for (Int_t iPart2 = 0; iPart2 < nPartc[1]; iPart2++)
            {
                if (timeLos[1][iPart2] > 0. && timeLos[0][iPart1] > 0.)
                {
                    fh_dt_los1_los2->Fill(timeLos[1][iPart2] - timeLos[0][iPart1]);
                    if ((timeLos[1][iPart2] - timeLos[0][iPart1]) > 2. &&
                        (timeLos[1][iPart2] - timeLos[0][iPart1] < 2.4))
                        fh_los_ToT_vs_ToT->Fill(totsum[0][iPart1], totsum[1][iPart2]);
                }
            }
        }
    }

    //----------------------------------------------------------------------
    // SCI8 detector
    //----------------------------------------------------------------------
    std::vector<double_t> time_first(fNofLosDetectors, -1.);

    Double_t timeS8_V[10][2], timeS8_L[10][2], timeS8_T[10][2], timeSci8M[10], Sci8TresM[10], timeSci8T[10],
        Sci8TresT[10], timeSci8[10], totsumS8[10], totS8[10][8]; // [multihit][pm]
    for (Int_t i = 0; i < 10; i++)
    {
        timeSci8M[i] = 0.0;
        Sci8TresM[i] = 0.0 / 0.0;
        timeSci8T[i] = 0.0;
        Sci8TresT[i] = 0.0 / 0.0;
        timeSci8[i] = 0.0;
        totsumS8[i] = 0.0;
        for (Int_t j = 0; j < 2; j++)
        {
            timeS8_L[i][j] = 0.0 / 0.0;
            timeS8_T[i][j] = 0.0 / 0.0;
            totS8[i][j] = 0.0 / 0.0;
        }
    }

    Int_t MultipS8;
    Int_t nPartS8;
    Int_t lhitmem[fNofLosDetectors] = { 0 };
    Double_t tof_mem[fNofLosDetectors];
    for (Int_t i = 0; i < fNofLosDetectors; i++)
    {
        tof_mem[i] = 0. / 0.;
    }

    if (fMappedItems.at(DET_SCI8))
    {
        auto det = fMappedItems.at(DET_SCI8);
        Int_t nHits = det->GetEntriesFast();

        MultipS8 = nHits;

        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BSci8MappedData* hit = (R3BSci8MappedData*)det->At(ihit);
            if (!hit)
                continue;

            // channel numbers are stored 1-based (1..n)
            Int_t iDet = hit->GetDetector(); // 1..
            Int_t iCha = hit->GetChannel();  // 1..

            fh_sci8_channels->Fill(iCha);
        }
    }

    if (fCalItems.at(DET_SCI8))
    {
        auto det = fCalItems.at(DET_SCI8);
        nPartS8 = det->GetEntriesFast();

        fh_sci8_multihit->Fill(nPartS8);

        Int_t iDet = 0;
        Int_t nPartS8_VFTX[2] = { 0 };
        Int_t nPartS8_LEAD[2] = { 0 };
        Int_t nPartS8_TRAI[2] = { 0 };

        for (Int_t iPart = 0; iPart < nPartS8; iPart++)
        {

            /*
             * nPart is the number of particle passing through Sci8 detector in one event
             */

            R3BSci8CalData* calDataS8 = (R3BSci8CalData*)det->At(iPart);
            iDet = calDataS8->GetDetector();

            // VFTX Channels 1-2:
            if (!(IS_NAN(calDataS8->fTimeV_r_ns)))
            {
                timeS8_V[iPart][0] = calDataS8->fTimeV_r_ns;
                nPartS8_VFTX[0] += 1;
            }
            if (!(IS_NAN(calDataS8->fTimeV_l_ns)))
            {
                timeS8_V[iPart][1] = calDataS8->fTimeV_l_ns;
                nPartS8_VFTX[1] += 1;
            }

            // TAMEX Channels 1-2:
            if (!(IS_NAN(calDataS8->fTimeL_r_ns)))
            {
                timeS8_L[iPart][0] = calDataS8->fTimeL_r_ns;
                nPartS8_LEAD[0] += 1;
            }
            if (!(IS_NAN(calDataS8->fTimeT_r_ns)))
            {
                timeS8_T[iPart][0] = calDataS8->fTimeT_r_ns;
                nPartS8_TRAI[0] += 1;
            }
            if (!(IS_NAN(calDataS8->fTimeL_l_ns)))
            {
                timeS8_L[iPart][1] = calDataS8->fTimeL_l_ns;
                nPartS8_LEAD[1] += 1;
            }
            if (!(IS_NAN(calDataS8->fTimeT_l_ns)))
            {
                timeS8_T[iPart][1] = calDataS8->fTimeT_l_ns;
                nPartS8_TRAI[1] += 1;
            }

            if (iPart > 0 && MultipS8 % 2 == 0)
            {
                for (int k = 0; k < 2; k++)
                {
                    if (timeS8_V[iPart][k] > 0. && timeS8_V[iPart - 1][k] > 0. && !(IS_NAN(timeS8_V[iPart][k])) &&
                        !(IS_NAN(timeS8_V[iPart - 1][k])))
                    {
                        fh_sci8_dt_hits->Fill(timeS8_V[iPart][k] - timeS8_V[iPart - 1][k]);
                    }
                    if (timeS8_L[iPart][k] > 0. && timeS8_L[iPart - 1][k] > 0. && !(IS_NAN(timeS8_L[iPart][k])) &&
                        !(IS_NAN(timeS8_L[iPart - 1][k])))
                    {
                        fh_sci8_dt_hits_l->Fill(timeS8_L[iPart][k] - timeS8_L[iPart - 1][k]);
                    }
                    if (timeS8_T[iPart][k] > 0. && timeS8_T[iPart - 1][k] > 0. && !(IS_NAN(timeS8_T[iPart][k])) &&
                        !(IS_NAN(timeS8_T[iPart - 1][k])))
                    {
                        fh_sci8_dt_hits_t->Fill(timeS8_T[iPart][k] - timeS8_T[iPart - 1][k]);
                    }
                }
            }

            if (iDet == 1)
            {

                // if(iPart >= 0)
                if (1 == 1)
                {
                    int nPMT = 0;
                    int nPMV = 0;
                    int ilc = iPart;

                    for (int ipm = 0; ipm < 2; ipm++)
                    {

                        if (timeS8_T[iPart][ipm] > 0. && timeS8_L[iPart][ipm] > 0. && !(IS_NAN(timeS8_T[iPart][ipm])) &&
                            !(IS_NAN(timeS8_L[iPart][ipm])))
                        {
                            while (timeS8_T[iPart][ipm] - timeS8_L[iPart][ipm] < 0.)
                            {
                                timeS8_T[iPart][ipm] = timeS8_T[iPart][ipm] + 2048. * fClockFreq;
                            }

                            nPMT = nPMT + 1;
                            totS8[iPart][ipm] = timeS8_T[iPart][ipm] - timeS8_L[iPart][ipm];
                        }

                        totsumS8[iPart] += totS8[iPart][ipm];

                        if (totS8[iPart][ipm] != 0. && !(IS_NAN(totS8[iPart][ipm])))
                            fh_sci8_tot->Fill(ipm + 1, totS8[iPart][ipm]);

                        if (timeS8_L[iPart][ipm] > 0. && !(IS_NAN(timeS8_L[iPart][ipm])))
                            timeSci8T[iPart] += timeS8_L[iPart][ipm];

                        if (timeS8_V[iPart][ipm] > 0. && !(IS_NAN(timeS8_V[iPart][ipm])))
                        {
                            timeSci8M[iPart] += timeS8_V[iPart][ipm];
                            nPMV = nPMV + 1;
                        }
                    }

                    totsumS8[iPart] = totsumS8[iPart] / nPMT;
                    /*
                    if(totsum[iPart] < 88.) cout<<fNEvents<<"; "<<nPart<<"; "<<iPart<<",
                    "<<totsum[iPart]<<tot[iPart][0]<<
                    ", "<<tot[iPart][1]<<", " <<tot[iPart][2]<<", "<<tot[iPart][3]<<", "<<tot[iPart][4]<<", "
                    <<tot[iPart][5]<<", "<<tot[iPart][6]<<", "<<tot[iPart][7]<<endl;
                    */
                    timeSci8M[iPart] = timeSci8M[iPart] / nPMV;
                    timeSci8T[iPart] = timeSci8T[iPart] / nPMT;

                    timeSci8[iPart] = timeSci8M[iPart];

                    Double_t tof = 0. / 0.;
                    // at first, take the first LOS-hit

                    if (!IS_NAN(timeLos[1][0]) && timeLos[1][0] > 0.)
                    {
                        if (timeSci8[iPart] - timeLos[1][0] < 8192. * 5. / 2.)
                            timeSci8[iPart] += 8192. * 5.;
                        if (timeSci8[iPart] - timeLos[1][0] > 8192. * 5. / 2.)
                            timeSci8[iPart] -= 8192. * 5.;
                        tof = timeSci8[iPart] - timeLos[1][0];
                    }

                    // cout<<"tof "<<tof<<"  "<<timeSci8[iPart]<<"  "<<timeLos[1][0]<<endl;
                    // loop over multihits of LOS
                    for (Int_t ilos = 0; ilos < fNofLosDetectors; ilos++)
                    {
                        for (Int_t lhit = 0; lhit < nPartc[1]; lhit++)
                        {
                            if (IS_NAN(timeLos[ilos][lhit]))
                                continue;
                            if (timeSci8[iPart] - timeLos[ilos][lhit] < 8192. * 5. / 2.)
                                timeSci8[iPart] += 8192. * 5.;
                            if (timeSci8[iPart] - timeLos[ilos][lhit] > 8192. * 5. / 2.)
                                timeSci8[iPart] -= 8192. * 5.;
                            Double_t tofi = timeSci8[iPart] - timeLos[ilos][lhit];
                            // if hit is closer to expected value then take it
                            if (abs(tofi - 7852.) < abs(tof - 7852.))
                            {
                                tof = tofi;
                                lhitmem[ilos] = lhit;
                                tof_mem[ilos] = tof;
                                // cout<<"tofi "<<tof<<"  "<<timeSci8[iPart]<<"  "<<timeLos[1][0]<<endl;
                            }
                        }

                        // cout<<"lhitmem "<<ilos<<", "<<lhitmem[ilos]<<", "<<tof_mem[ilos]<<endl;

                        if (ilos == 1)
                            fh_tof_sci8->Fill(tof - 7852.);
                    }

                    if (nPMV == 2)
                        Sci8TresM[iPart] = (timeS8_V[iPart][1] - timeS8_V[iPart][0]);
                    if (nPMT == 2)
                        Sci8TresT[iPart] = (timeS8_L[iPart][1] - timeS8_L[iPart][0]);

                    if (nPMV == 2)
                        fh_sci8_tres_MCFD->Fill(Sci8TresM[iPart]);
                    if (nPMT == 2)
                        fh_sci8_tres_TAMEX->Fill(Sci8TresT[iPart]);
                    if (nPMT == 2)
                        fh_sci8_tot_mean->Fill(totsumS8[iPart]);
                }
            }
            else
            {
                cout << "Wrong detector ID for Sci8!" << endl;
            }
        }

        for (int ik = 0; ik < 2; ik++)
        {
            fh_sci8_multihitVFTX->Fill(ik + 1, nPartS8_VFTX[ik]);
            fh_sci8_multihitLEAD->Fill(ik + 1, nPartS8_LEAD[ik]);
            fh_sci8_multihitTRAI->Fill(ik + 1, nPartS8_TRAI[ik]);
        }
    }

    //----------------------------------------------------------------------
    // Fiber detectors
    //----------------------------------------------------------------------
    Double_t dtime = 0.0 / 0.0;
    Bool_t fib1a_cut = false;
    Bool_t fib1b_cut = false;
    Double_t yfib, xfib, tfibMax = 0. / 0.;
    Double_t rfib = 0.0 / 0.0;
    Bool_t LosFibCorr = false;
    Bool_t los_fib_cor[NOF_FIB_DET];

    for (Int_t ifibcount = 0; ifibcount < NOF_FIB_DET; ifibcount++)
    {
        los_fib_cor[ifibcount] = false;
        Int_t iFib = 0;

        auto detMapped = fMappedItems.at(DET_FI_FIRST + ifibcount);
        auto detHit = fHitItems.at(DET_FI_FIRST + ifibcount);

        if (detMapped)
        {
            Int_t nHits = detMapped->GetEntriesFast();
            std::vector<UInt_t> mapmt_num(512);
            std::vector<UInt_t> spmt_num(16);
            for (Int_t ihit = 0; ihit < nHits; ihit++)
            {
                R3BBunchedFiberMappedData* hit = (R3BBunchedFiberMappedData*)detMapped->At(ihit);
                if (!hit)
                    continue;

                // channel numbers are stored 1-based (1..n)
                Int_t iCha = hit->GetChannel(); // 1..

                if (hit->IsMAPMT() && hit->IsLeading())
                {
                    fh_channels_Fib[ifibcount]->Fill(iCha); // Fill which clockTDC channel has events
                    ++mapmt_num.at(hit->GetChannel() - 1);  // multihit of a given clockTDC channel
                }

                if (!hit->IsMAPMT() && hit->IsLeading())
                {
                    fh_channels_single_Fib[ifibcount]->Fill(iCha); // Fill which single PMT channel has events
                    ++spmt_num.at(hit->GetChannel() - 1);          // multihit of a given PADI channel
                }
            }

            for (int i = 0; i < 512; ++i)
            {
                auto m = mapmt_num.at(i);
                if (m > 0)
                    fh_multihit_m_Fib[ifibcount]->Fill(i + 1, m); // multihit of a given clockTDC channel
            }

            for (int i = 0; i < 16; ++i)
            {
                auto s = spmt_num.at(i);
                if (s > 0)
                    fh_multihit_s_Fib[ifibcount]->Fill(i + 1, s); // multihit of a given PADI channel
            }
        }

        if (detHit)
        {
            Int_t nHits = detHit->GetEntriesFast();
            Int_t iFibMax = -100;
            Int_t iFibMax_MA = -100;
            Double_t xpos = 0. / 0.;
            Double_t ypos = 0. / 0.;
            Double_t totMax = 0.;
            Double_t tofMax;
            Double_t tfib = 0. / 0., tof_fib = 0. / 0., tof_fib_s = 0. / 0., tof_fib_m = 0. / 0.;
            Double_t randx;
            Double_t totMax_MA = 0.;
            Double_t spmtMax;
            Double_t mapmtMax;
            Double_t posfib = 0. / 0.;
            Double_t totMax_S = 0.;
            Double_t yposMax = -10000;
            Double_t tof_fibMax = -10000;

            for (Int_t ihit = 0; ihit < nHits; ihit++)
            {
                Double_t tMAPMT = 0. / 0.;
                Double_t tSPMT = 0. / 0.;

                R3BBunchedFiberHitData* hit = (R3BBunchedFiberHitData*)detHit->At(ihit);
                if (!hit)
                    continue;

                iFib = hit->GetFiberId(); // 1..

                // times
                tMAPMT = hit->GetMAPMTTime_ns();
                tSPMT = hit->GetSPMTTime_ns();
                /*
                                dtime = tMAPMT-tSPMT;

                                // "Push" the Fib times in the same cycle with LOS:
                                // Los coarse counter has 8192 channels with 5 ns each
                                // SPMT coarse counter has 2048 channels with 5 ns each
                                if(timeLos[0][0]>0. && timeLos[0][0] < 8192.*5. && !(IS_NAN(timeLos[0][0])))
                                {
                                    while(tSPMT - timeLos[0][0] < 2048.*5./2.)
                                    {
                                        tSPMT = tSPMT + 2048.*5.;
                                    }
                                    while(tSPMT - timeLos[0][0] > 2048.*5./2.)
                                    {
                                        tSPMT = tSPMT - 2048.*5.;
                                    }
                                }

                                // "Push" two times in the same clock cycle:
                                // MAPMT coarse counter has 4096 channels with 1000/150=6.67 ns each
                                // SPMT coarse counter has 2048 channels with 5 ns each

                                dtime = tMAPMT-tSPMT;

                                while(dtime > 1024./2.*1000./150.)
                                {
                                    tMAPMT = tMAPMT - 1024.*1000./150.;
                                    dtime = tMAPMT - tSPMT;
                                }
                                while(dtime < -1024./2.*1000./150.)
                                {
                                    tMAPMT = tMAPMT + 1024.*1000./150.;
                                    dtime = tMAPMT - tSPMT;
                                }

                                // Not-calibrated ToF:
                                tfib = (tMAPMT + tSPMT) / 2.;
                               // tfib = tSPMT;
                                if(tfib > 0. && !(IS_NAN(tfib)) ){
                                 for(Int_t imlos = 0; imlos < nPartc[0]; imlos++){
                                  if((tfib - timeLos[0][imlos]) > 2575. && (tfib - timeLos[0][imlos]) < 2600)  tof_fib =
                tfib - timeLos[0][imlos];
                                  }
                                 }
                               // Not-calibrated position:
                                xpos=iFib;
                                ypos=dtime;

                                // find fiber with most light which has resonable ToF and y-position
                                Double_t ToT_MA = hit->GetMAPMTToT_ns();
                                if(ypos>70. && ypos<100.){
                                    if(ToT_MA>totMax_MA ) {
                                        totMax_MA=ToT_MA;
                                        iFibMax_MA=iFib;
                                        tfibMax = tfib;
                                    }
                                }

                                if(hit->GetSPMTToT_ns() > 0)
                                {
                                    fh_fibers_Fib[ifibcount]->Fill(iFib);
                                    fh_ToT_s_Fib[ifibcount]->Fill(iFib,hit->GetSPMTToT_ns());
                                    fh_ToT_m_Fib[ifibcount]->Fill(iFib,hit->GetMAPMTToT_ns());
                                    fh_time_Fib[ifibcount]->Fill(iFib,tMAPMT-tSPMT);
                                    fh_Fib_ToF[ifibcount]->Fill(iFib,tof_fib);
                                    fh_Fib_pos[ifibcount]->Fill(xpos,ypos);
                                    fh_Fib_vs_Events[ifibcount]->Fill(fNEvents,iFib);
                                    if(ifibcount==12 || ifibcount==13){
                                        fh_ToT_single_Fib[ifibcount]->Fill((iFib-1)%2+1+2*((iFib-1)/512),hit->GetSPMTToT_ns());
                //                    cout<<"Test: "<<ifibcount<<" ifib: "<<iFib<<" single PMT: "<<
                (iFib)%2+1+2*((iFib-1)/512)<<endl;
                                    }

                                    if(ifibcount==14 || ifibcount==15){
                                        fh_ToT_single_Fib[ifibcount]->Fill((iFib-1)%2+1+2*((iFib-1)/512),hit->GetSPMTToT_ns());
                //                    cout<<"Test: "<<ifibcount<<" ifib: "<<iFib<<" single PMT: "<<
                (iFib-1)%2+1+2*((iFib-1)/512)<<endl;
                                    }

                                    if(ifibcount==9 || ifibcount==10 ){
                                        fh_ToT_single_Fib[ifibcount]->Fill((iFib-1)/256+1,hit->GetSPMTToT_ns());
                                    }

                                    if(ifibcount==4 || ifibcount==5 ){
                                        fh_ToT_single_Fib[ifibcount]->Fill((iFib-1)%2+1,hit->GetSPMTToT_ns());
                                    }
                                    if(ifibcount==0 || ifibcount==1 ){
                                        fh_ToT_single_Fib[ifibcount]->Fill(1,hit->GetSPMTToT_ns());
                                    }

                                }
                */
                // if not both times present, go to next
                if (!(tMAPMT > 0.) || !(tSPMT > 0.))
                    continue;

                // loop over all hits of LOS1:
                for (Int_t lhit = 0; lhit < nPartc[1]; lhit++)
                {
                    if (IS_NAN(timeLos[1][lhit]))
                        continue;

                    // if not multiplicity 8 or pile-up, go to next event
                    if (!iLOSType[1][lhit])
                        continue;

                    // "Push" the Fib times in the same cycle with LOS:
                    // Los coarse counter has 8192 channels with 5 ns each
                    // SPMT coarse counter has 2048 channels with 5 ns each
                    if (timeLos[1][lhit] > 0. && timeLos[1][lhit] < 8192. * 5. && !(IS_NAN(timeLos[1][lhit])))
                    {
                        while (tSPMT - timeLos[1][lhit] < 2048. * 5. / 2.)
                        {
                            tSPMT = tSPMT + 2048. * 5.;
                        }
                        while (tSPMT - timeLos[1][lhit] > 2048. * 5. / 2.)
                        {
                            tSPMT = tSPMT - 2048. * 5.;
                        }
                    }

                    // "Push" two times in the same clock cycle:
                    // MAPMT coarse counter has 4096 channels with 1000/150=6.67 ns each
                    // SPMT coarse counter has 2048 channels with 5 ns each
                    dtime = tMAPMT - tSPMT;
                    while (dtime > 1024. / 2. * 1000. / 150.)
                    {
                        tMAPMT = tMAPMT - 1024. * 1000. / 150.;
                        dtime = tMAPMT - tSPMT;
                    }
                    while (dtime < -1024. / 2. * 1000. / 150.)
                    {
                        tMAPMT = tMAPMT + 1024. * 1000. / 150.;
                        dtime = tMAPMT - tSPMT;
                    }

                    // Not-calibrated ToF:
                    tfib = (tMAPMT + tSPMT) / 2.;
                    // tfib = tSPMT;
                    // tfib = tMAPMT;
                    if (tfib > 0. && !(IS_NAN(tfib)) && timeLos[1][lhit] > 0. && !(IS_NAN(timeLos[1][lhit])))
                        tof_fib = tfib - timeLos[1][lhit];

                    // if ToF between LOS1 and fiber is not okay, go to next
                    if (tof_fib < tof_fiber[ifibcount] - 10. || tof_fib > tof_fiber[ifibcount] + 10.)
                        continue;

                    // Not-calibrated position:
                    randx = (std::rand() / (float)RAND_MAX);
                    //                if(iFib > 0) xpos = ((-n_fiber[ifibcount]/2.+iFib+(0.5-randx)))*0.21;
                    xpos = iFib;
                    ypos = dtime;

                    // if not resonable y-position, go to next
                    if (ypos < 70. || ypos > 100.)
                        continue;

                    // find fiber with most light which has resonable ToF and y-position
                    Double_t ToT_MA = hit->GetMAPMTToT_ns();
                    if (ToT_MA > totMax_MA)
                    {
                        totMax_MA = ToT_MA;
                        iFibMax_MA = iFib;
                        totMax_S = hit->GetSPMTToT_ns();
                        yposMax = ypos;
                        tof_fibMax = tof_fib;

                        // cout<<"new imax "<<iFibMax_MA<<" max "<<totMax_MA<<endl;
                    }
                    /*
                                        fh_fibers_Fib[ifibcount]->Fill(iFib);
                                        fh_ToT_s_Fib[ifibcount]->Fill(iFib,hit->GetSPMTToT_ns());
                                        fh_ToT_m_Fib[ifibcount]->Fill(iFib,hit->GetMAPMTToT_ns());
                                        fh_time_Fib[ifibcount]->Fill(iFib,tMAPMT-tSPMT);
                                        fh_Fib_ToF[ifibcount]->Fill(iFib,tof_fib);
                                        fh_Fib_pos_xy[ifibcount]->Fill(xpos,ypos);
                                        fh_Fib_vs_Events[ifibcount]->Fill(fNEvents,iFib);
                                        fh_ToT_single_Fib[ifibcount]->Fill(1,hit->GetSPMTToT_ns());
                    */
                } // end for (lhit)
            }     // end for(ihit)

            // LOS correlated to ifibMax
            if (tof_fibMax > tof_fiber[ifibcount] - 10. && tof_fib < tof_fiber[ifibcount] + 10.)
                los_fib_cor[ifibcount] = true;

            randx = (std::rand() / (float)RAND_MAX);
            if (iFibMax_MA > 0)
                posfib = (-n_fiber[ifibcount] / 2. + iFibMax_MA + (0.5 - randx)) * 0.221 / 10.; // in cm

            if (ifibcount == 0)
                yfib = posfib;
            if (ifibcount == 1)
                xfib = posfib;

            //	if(iFibMax_MA > 0)cout<<"FIB POS: "<<fNEvents<<", "<<ifibcount<<"; "<<iFibMax_MA<<",
            //"<<n_fiber[ifibcount]<<"; "<<posfib<<endl;

            // if(ifibcount==0 && iFibMax%32 == 0) fib1a_cut = true;
            // if(ifibcount==1 && iFibMax%32 == 0) fib1b_cut = true;

            // if(ifibcount==0 && (iFibMax == 100 || iFibMax == 200)) fib1a_cut = true;
            // if(ifibcount==1 && (iFibMax == 100 || iFibMax == 200)) fib1b_cut = true;

            // For position calibration:
            /*
                   if (ifibcount==0  && ((iFibMax_MA >68 && iFibMax_MA<88) || (iFibMax_MA >168 && iFibMax_MA<188))){
                    fib1a_cut=true;
                }
                   if (ifibcount==1  && ((iFibMax_MA >68 && iFibMax_MA<88) || (iFibMax_MA >168 && iFibMax_MA<188))){
                    fib1b_cut=true;
                }
            */
            // Full fib1a, 1b:

            if (ifibcount == 0 && (iFibMax_MA > 1 && iFibMax_MA < 265))
            {
                fib1a_cut = true;
            }
            if (ifibcount == 1 && (iFibMax_MA > 1 && iFibMax_MA < 265))
            {
                fib1b_cut = true;
            }

            // Center of LOS detector:
            /*
                if (ifibcount==0  && (iFibMax_MA >110 && iFibMax_MA<146)){
                    fib1a_cut=true;
                }
                   if (ifibcount==1  && (iFibMax_MA >110 && iFibMax_MA<146)){
                    fib1b_cut=true;
                }
              */
            if (nHits > 0)
                fh_mult_Fib[ifibcount]->Fill(nHits);

        } // end if(aHit[ifibcount])
    }     // end for(ifibcount)

    // LOS1 and fib events are correlated:
    if (los_fib_cor[0] && los_fib_cor[1])
        LosFibCorr = true;

    rfib = sqrt(xfib * xfib + yfib * yfib);

    if (foptcond == 1 && (fib1a_cut == true && fib1b_cut == true && rfib < 2.) && twin)
    {
        for (Int_t ilos = 0; ilos < fNofLosDetectors; ilos++)
        {

            for (Int_t imultlos = 0; imultlos < nPartc[ilos]; imultlos++)
            {
                if (iLOSType[ilos][imultlos] && nPartc[ilos] == 1)
                {
                    fh_los_pos_MCFD[ilos]->Fill(xV_cm[ilos][imultlos], yV_cm[ilos][imultlos]);
                    fh_los_pos_TAMEX[ilos]->Fill(xT_cm[ilos][imultlos], yT_cm[ilos][imultlos]);
                    fh_los_pos_ToT[ilos]->Fill(xToT_cm[ilos][imultlos], yToT_cm[ilos][imultlos]);
                    fh_los_tres_MCFD[ilos]->Fill(LosTresV[ilos][imultlos]);
                    fh_los_tres_TAMEX[ilos]->Fill(LosTresT[ilos][imultlos]);
                    // fh_los_tot_mean[ilos]->Fill(totsum[ilos][imultlos]);
                    for (int ipm = 0; ipm < 8; ipm++)
                    {
                        fh_los_tot[ilos]->Fill(ipm + 1, tot[ilos][imultlos][ipm]);
                    }
                }
            }
        }
        if (fCalItems.at(DET_LOS) && fNofLosDetectors == 2)
        {
            for (Int_t iPart1 = 0; iPart1 < nPartc[0]; iPart1++)
            {
                for (Int_t iPart2 = 0; iPart2 < nPartc[1]; iPart2++)
                {
                    if ((timeLos[1][iPart2] - timeLos[0][iPart1]) > 2.2 &&
                        (timeLos[1][iPart2] - timeLos[0][iPart1]) < 2.45 && timeLos[1][iPart2] > 0. &&
                        timeLos[0][iPart1] > 0.)
                    {
                        // fh_dt_los1_los2->Fill(timeLos[1][iPart2]-timeLos[0][iPart1]);
                        //  fh_los_ToT_vs_ToT->Fill(totsum[0][iPart1],totsum[1][iPart2]);
                    }
                }
            }
        }
    }

    if (fCalItems.at(DET_LOS) && LosFibCorr)
    {
        Int_t nevmem = 0;
        Double_t rpm = 4.75;
        Double_t lpm = 2. / 2;
        Double_t apm = 22.5 * 3.14159 / 180.;
        Double_t labs = 10.;
        Double_t anglepm[fNofLosDetectors][32][8];
        Double_t xpm[fNofLosDetectors][32][8];
        Double_t ypm[fNofLosDetectors][32][8];
        Double_t hpm[fNofLosDetectors][32][8];
        Double_t alpha[fNofLosDetectors][32][8];
        Double_t tot_corr[fNofLosDetectors][32][8];
        Double_t sum_alpha[fNofLosDetectors][32];
        Double_t totsum_corr[fNofLosDetectors][32];
        Double_t rtot[fNofLosDetectors][32];
        Double_t tot_lim[fNofLosDetectors] = { 0 };
        Double_t par0[8], par1[8], par2[8], par3[8], par4[8], par5[8];

        par0[0] = 1.61580e+02;  // 3.83810e+01;
        par1[0] = -1.17944e+00; //-1.09494e+00 ;
        par2[0] = 1.41921e+02;  // 1.28720e+02;
        par3[0] = -8.11592e+00; //-7.43137e+00;
        par4[0] = 0.;
        par5[0] = 1;

        par0[1] = 3.71160e+02;  // 1.02121e+02;
        par1[1] = -1.03041e+00; //-3.12171e+00;
        par2[1] = 1.32344e+02;  // 6.44474e+01;
        par3[1] = -9.86918e+00; //-1.87164e+01;
        par4[1] = 0.;
        par5[1] = 1.;

        par0[2] = 2.17270e+02;  // 1.24749e+02;
        par1[2] = -1.28556e+00; //-2.50991e+00;
        par2[2] = 1.51548e+02;  // 6.88595e+01;
        par3[2] = -8.29751e+00; //-1.43484e+01;
        par4[2] = 0.;
        par5[2] = 1.;

        par0[3] = 8.41271e+02;  // 6.55205e+01;
        par1[3] = -8.57171e-01; //-3.66726e+00;
        par2[3] = 1.79550e+02;  // 1.00290e+02;
        par3[3] = -7.47610e+00; //-7.36581e+00;
        par4[3] = 0.;
        par5[3] = 1.;

        par0[4] = 5.23367e+02;  // 5.30274e+01;
        par1[4] = -1.11625e+00; //-8.92102e+00;
        par2[4] = 1.58305e+02;  // 1.14336e+02;
        par3[4] = -7.60637e+00; //-4.33577e+00;
        par4[4] = 0.;
        par5[4] = 1.;

        par0[5] = 1.07023e+02;  // 5.15765e+01;
        par1[5] = -8.13576e-01; //-1.74148e+01;
        par2[5] = 2.27580e+02;  // 1.29832e+02;
        par3[5] = -5.51671e+00; //-3.40990e+00;
        par4[5] = 0.;
        par5[5] = 1.;

        par0[6] = 1.35075e+02;  // 4.20388e+01;
        par1[6] = -7.97772e+00; //-1.25505e+00;
        par2[6] = 2.10053e+02;  // 1.29497e+02;
        par3[6] = -1.13020e+00; //-6.92512e+00;
        par4[6] = 0.;
        par5[6] = 1.;

        par0[7] = 1.62558e+02;  // 2.02182e+01;
        par1[7] = -1.70696e+00; //-1.11835e+00;
        par2[7] = 1.22895e+02;  // 1.42994e+02;
        par3[7] = -1.10376e+01; //-6.83987e+00;
        par4[7] = 0.;
        par5[7] = 1.;

        for (Int_t i = 0; i < fNofLosDetectors; i++)
        {
            for (Int_t k = 0; k < 32; k++)
            {
                sum_alpha[i][k] = 0.;
                totsum_corr[i][k] = 0.;
                rtot[i][k] = 0. / 0.;
            }
        }
        Double_t rimpact = 0. / 0.;

        rimpact = sqrt(xfib * xfib + yfib * yfib);

        for (Int_t idet = 0; idet < fNofLosDetectors; idet++)
        {
            if (nPartc[idet] == 1)
            {
                for (Int_t imul = 0; imul < nPartc[idet]; imul++)
                {
                    if (iLOSType[idet][imul])
                    {
                        rtot[idet][imul] =
                            sqrt(xToT_cm[idet][imul] * xToT_cm[idet][imul] + yToT_cm[idet][imul] * yToT_cm[idet][imul]);
                        for (Int_t ipm = 0; ipm < 8; ipm++)
                        {
                            if (tot[idet][imul][ipm] > 0. && !(IS_NAN(tot[idet][imul][ipm])) && !(IS_NAN(xfib)) &&
                                !(IS_NAN(yfib)))
                            {
                                anglepm[idet][imul][ipm] = apm + 2. * (7. - ipm) * apm;
                                xpm[idet][imul][ipm] = rpm * sin(anglepm[idet][imul][ipm]);
                                ypm[idet][imul][ipm] = rpm * cos(anglepm[idet][imul][ipm]);
                                hpm[idet][imul][ipm] =
                                    sqrt((xpm[idet][imul][ipm] - xfib) * (xpm[idet][imul][ipm] - xfib) +
                                         (ypm[idet][imul][ipm] - yfib) * (ypm[idet][imul][ipm] - yfib));
                                // alpha[idet][imul][ipm] = 2. * atan(lpm/hpm[idet][imul][ipm]);
                                alpha[idet][imul][ipm] =
                                    2. * lpm * (1. - lpm / hpm[idet][imul][ipm]) / (hpm[idet][imul][ipm] - lpm);
                                sum_alpha[idet][imul] = sum_alpha[idet][imul] + alpha[idet][imul][ipm];
                                // if(idet == 0) tot_corr[idet][imul][ipm] =
                                // tot[idet][imul][ipm]/alpha[idet][imul][ipm]*exp(-hpm[idet][imul][ipm]/10.);
                                //  if(idet == 1){

                                tot_corr[idet][imul][ipm] = 100. * tot[idet][imul][ipm] /
                                                            (par0[ipm] * exp(hpm[idet][imul][ipm] / par1[ipm]) +
                                                             par2[ipm] * exp(hpm[idet][imul][ipm] / par3[ipm]) +
                                                             par4[ipm] * exp(hpm[idet][imul][ipm] / par5[ipm]));
                                if (hpm[idet][imul][ipm] < 2.2 || hpm[idet][imul][ipm] > 6.5)
                                    tot_corr[idet][imul][ipm] = 0. / 0.;

                                if (idet == 0)
                                    fh_test_corr[ipm]->Fill(hpm[idet][imul][ipm], tot_corr[idet][imul][ipm]);
                                // }

                                // tot_corr[idet][imul][ipm] =
                                // tot[idet][imul][ipm]/TMath::Power(alpha[idet][imul][ipm],0.75);

                                totsum_corr[idet][imul] += tot_corr[idet][imul][ipm];
                                nevmem = fNEvents;

                                // cout<<idet<<", "<<ipm<<", "<<hpm[idet][ipm]<<endl;

                                if (idet == 0)
                                {
                                    fh_test[ipm]->Fill(hpm[idet][imul][ipm], tot[idet][imul][ipm]);
                                }
                            }
                        }

                        /*
                         totsum_corr[idet][imul] = (sqrt(tot[idet][imul][0]*tot[idet][imul][4])+
                                                    sqrt(tot[idet][imul][1]*tot[idet][imul][5])+
                                                    sqrt(tot[idet][imul][2]*tot[idet][imul][6])+
                                                    sqrt(tot[idet][imul][3]*tot[idet][imul][7]))/4.;
                        */
                        totsum_corr[idet][imul] = totsum_corr[idet][imul] / 8.;
                        sum_alpha[idet][imul] = sum_alpha[idet][imul] * 180. / 3.14159;

                        if (idet == 1)
                            tot_lim[idet] = -3.e-6 * dE_TWIN * dE_TWIN + 0.03 * dE_TWIN + 25.172;
                        if (idet == 0)
                            tot_lim[idet] = -8e-7 * dE_TWIN * dE_TWIN + 0.0169 * dE_TWIN + 27.399;

                        if (rfib < 3. && totsum[idet][imul] > 0. && twin)
                        {
                            fh_los_tot_mean_corr[idet]->Fill(totsum_corr[idet][imul]);
                            fh_los_tot_mean[idet]->Fill(totsum[idet][imul]);
                            fh_los_ihit_ToT[idet]->Fill(rimpact, totsum_corr[idet][imul]);
                            fh_los_rw[idet]->Fill(sum_alpha[idet][imul]);
                            fh_los_twin[idet]->Fill(totsum_corr[idet][imul], dE_TWIN);

                            /* if(imul == lhitmem[idet])*/ fh_los_tot_S8[idet]->Fill(tof_mem[idet] - 7852.64,
                                                                                     totsum_corr[idet][imul]);
                        }

                        // Double_t tot_test = totsum_corr[idet][imul] * 360./sum_alpha[idet][imul];

                        // if(fNEvents == nevmem) cout<<"Output "<<totsum_corr[idet]<<", "<<totsum[idet][0]<<",
                        // "<<sum_alpha[idet]<<endl;
                    }
                }
            }
        }
    }
    //----------------------------------------------------------------------
    // TOFD
    //----------------------------------------------------------------------

    Int_t NumPaddles[4] = { 0 };
    if (fMappedItems.at(DET_TOFD))
    {
        auto det = fMappedItems.at(DET_TOFD);
        Int_t nMapped = det->GetEntriesFast();
        Int_t iPlaneMem = 1, iBarMem = 0;
        for (Int_t imapped = 0; imapped < nMapped; imapped++)
        {
            auto mapped = (R3BTofdMappedData const*)det->At(imapped);
            if (!mapped)
                continue; // should not happen

            Int_t const iPlane = mapped->GetDetectorId(); // 1..n
            Int_t const iBar = mapped->GetBarId();        // 1..n
            Int_t const iSide = mapped->GetSideId();      // 1..n
            Int_t const iEdge = mapped->GetEdgeId();

            if (iPlaneMem != iPlane)
                iBarMem = 0;
            if (iPlane == 1 && iBarMem != iBar && iSide == 1 && iEdge == 1)
                NumPaddles[0] += 1;
            if (iPlane == 2 && iBarMem != iBar && iSide == 1 && iEdge == 1)
                NumPaddles[1] += 1;
            if (iPlane == 3 && iBarMem != iBar && iSide == 1 && iEdge == 1)
                NumPaddles[2] += 1;
            if (iPlane == 3 && iBarMem != iBar && iSide == 1 && iEdge == 1)
                NumPaddles[3] += 1;

            //  cout<<imapped<<", "<<iPlane<<"; "<<iPlaneMem<<"; "<<iBar<<"; "<<iBarMem<<",
            //  "<<NumPaddles[iPlane-1]<<endl;

            iPlaneMem = iPlane;
            iBarMem = iBar;

            if (iPlane <= fNofPlanes)
            {
                if (iSide == 1)
                    fh_tofd_channels[iPlane - 1]->Fill(-iBar - 1);
                if (iSide == 2)
                    fh_tofd_channels[iPlane - 1]->Fill(iBar);
            }
        }
    }

    if (fCalItems.at(DET_TOFD))
    {
        auto det = fCalItems.at(DET_TOFD);
        Int_t nCals = det->GetEntriesFast();

        Double_t tot1[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t tot2[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t t_paddle[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t t1l[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t t2l[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t t1t[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t t2t[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };
        Double_t ToF[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0. / 0. };

        Bool_t Bar_present[10][N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { false };

        Int_t iBarMem = 0;
        Int_t jmult[N_PLANE_MAX_TOFD][N_PADDLE_MAX_TOFD] = { 0 };

        unsigned long long time0 = header->GetTimeStamp();
        double_t time1 = -1.;

        for (Int_t ical = 0; ical < nCals; ical++)
        {
            auto cal = (R3BTofdCalData const*)det->At(ical);
            if (!cal)
                continue; // should not happen

            Int_t const iPlane = cal->GetDetectorId(); // 1..n
            Int_t const iBar = cal->GetBarId();        // 1..n

            //      std::cout << iPlane << ' ' << iBar <<
            //          ',' << cal->GetTimeBL_ns() <<
            //          ' ' << cal->GetTimeBT_ns() <<
            //          ' ' << cal->GetTimeTL_ns() <<
            //          ' ' << cal->GetTimeTT_ns() << std::endl;

            // get all times of one bar
            /*
                         cout<<"TOFD Online: "<<fNEvents<<", "<<nCals<<"; "<<ical<<"; "<<iPlane<<", "<<iBar<<",
               "<<iBarMem<<", "<<", "<< cal->fTime1L_ns<<", "<<cal->fTime1T_ns<<", "<< cal->fTime2L_ns<<",
               "<<cal->fTime2T_ns<<endl;
            */
            /*
                  if(iBar != iBarMem )
                  {
                jmult[iPlane-1][iBar-1] = 0;
                if (!IS_NAN(cal->GetTimeBL_ns())) t1l[0][iPlane-1][iBar-1] = cal->GetTimeBL_ns();
                if (!IS_NAN(cal->GetTimeBT_ns())) t1t[0][iPlane-1][iBar-1] = cal->GetTimeBT_ns();
                if (!IS_NAN(cal->GetTimeTL_ns())) t2l[0][iPlane-1][iBar-1] = cal->GetTimeTL_ns();
                if (!IS_NAN(cal->GetTimeTT_ns())) t2t[0][iPlane-1][iBar-1] = cal->GetTimeTT_ns();
                Bar_present[0][iPlane-1][iBar-1] = true;
                  }
                  else
                  {
                jmult[iPlane-1][iBar-1] = jmult[iPlane-1][iBar-1] + 1;
                Int_t jm = jmult[iPlane-1][iBar-1];
                if (!IS_NAN(cal->GetTimeBL_ns())) t1l[jm][iPlane-1][iBar-1] = cal->GetTimeBL_ns();
                if (!IS_NAN(cal->GetTimeBT_ns())) t1t[jm][iPlane-1][iBar-1] = cal->GetTimeBT_ns();
                if (!IS_NAN(cal->GetTimeTL_ns())) t2l[jm][iPlane-1][iBar-1] = cal->GetTimeTL_ns();
                if (!IS_NAN(cal->GetTimeTT_ns())) t2t[jm][iPlane-1][iBar-1] = cal->GetTimeTT_ns();
                Bar_present[jm][iPlane-1][iBar-1] = true;
                  }
                  iBarMem = iBar;
            */

            Int_t jm = jmult[iPlane - 1][iBar - 1];
            /*		if (!IS_NAN(cal->GetTimeBL_ns())) t1l[jm][iPlane-1][iBar-1] = cal->GetTimeBL_ns();
                    if (!IS_NAN(cal->GetTimeBT_ns())) t1t[jm][iPlane-1][iBar-1] = cal->GetTimeBT_ns();
                    if (!IS_NAN(cal->GetTimeTL_ns())) t2l[jm][iPlane-1][iBar-1] = cal->GetTimeTL_ns();
                    if (!IS_NAN(cal->GetTimeTT_ns())) t2t[jm][iPlane-1][iBar-1] = cal->GetTimeTT_ns();*/
            Bar_present[jm][iPlane - 1][iBar - 1] = true;
            jmult[iPlane - 1][iBar - 1] = jmult[iPlane - 1][iBar - 1] + 1;

            double_t time2 = 0.;
            if (time1 < 0 && iPlane == 2)
                time1 = (t1l[jm][iPlane - 1][iBar - 1] + t2l[jm][iPlane - 1][iBar - 1]) / 2.;
            if (iPlane == 2)
                time2 = (t1l[jm][iPlane - 1][iBar - 1] + t2l[jm][iPlane - 1][iBar - 1]) / 2.;
            if (time1 > 0. && time2 > 0. && time2 > time1)
            {
                // cout<<"Time Test "<<time0<<"  "<<time1<< "   "<< time2 <<"  " <<time_previous_event <<endl;
                fh_TimePreviousEvent->Fill(time2 - time1);
                time2 = time1;
            }
        }

        /*
                for (Int_t ipl = 0; ipl < N_PLANE_MAX_TOFD; ipl++)
                for(Int_t ibr = 0; ibr < N_PADDLE_MAX_TOFD; ibr++)
                {
                fh_tofd_multihit[ipl]->Fill(ibr+1,jmult[ipl][ibr]+1);
                }
         */
        for (Int_t ipl = 0; ipl < N_PLANE_MAX_TOFD; ipl++)
            for (Int_t ibr = 0; ibr < N_PADDLE_MAX_TOFD; ibr++)
                for (Int_t jm = 0; jm < jmult[ipl][ibr]; jm++)
                {

                    if (Bar_present[jm][ipl][ibr])
                    {
                        Int_t iPlane = ipl + 1; // 1..n
                        Int_t iBar = ibr + 1;   // 1..n

                        fh_tofd_multihit[ipl]->Fill(ibr + 1, jmult[ipl][ibr]);

                        // calculate time over threshold and check if clock counter went out of range

                        while (t1t[jm][iPlane - 1][iBar - 1] < t1l[jm][iPlane - 1][iBar - 1])
                        {
                            t1t[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                        }

                        while (t2t[jm][iPlane - 1][iBar - 1] < t2l[jm][iPlane - 1][iBar - 1])
                        {
                            t2t[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                        }

                        int ilc = jm;

                        if (!(IS_NAN(timeLos[0][ilc])) && timeLos[0][ilc] > 0.)
                        {

                            while (t1l[jm][iPlane - 1][iBar - 1] < timeLos[0][ilc])
                            {
                                t1t[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                                t1l[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                            }
                            while (t2l[jm][iPlane - 1][iBar - 1] < timeLos[0][ilc])
                            {
                                t2t[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                                t2l[jm][iPlane - 1][iBar - 1] += 2048. * fClockFreq;
                            }
                        }

                        // ToF
                        t_paddle[jm][iPlane - 1][iBar - 1] =
                            (t1l[jm][iPlane - 1][iBar - 1] + t2l[jm][iPlane - 1][iBar - 1]) / 2.;

                        //  between 2 bars in 2 planes
                        if (ipl > 0)
                        {
                            fh_tofd_dt[ipl - 1]->Fill(iBar,
                                                      t_paddle[jm][ipl][iBar - 1] - t_paddle[jm][ipl - 1][iBar - 1]);
                        }

                        if (!(IS_NAN(timeLos[0][ilc])) && timeLos[0][ilc] > 0.)
                        {
                            // between LOS and paddle
                            ToF[jm][iPlane - 1][iBar - 1] =
                                fmod(t_paddle[jm][iPlane - 1][iBar - 1] - timeLos[0][ilc] + 5 * 8192, 5 * 2048);
                            fh_tofd_ToF[iPlane - 1]->Fill(iBar, ToF[jm][iPlane - 1][iBar - 1]);
                        }
                        // ToT
                        tot1[jm][iPlane - 1][iBar - 1] = t1t[jm][iPlane - 1][iBar - 1] - t1l[jm][iPlane - 1][iBar - 1];
                        if (tot1[jm][iPlane - 1][iBar - 1] < 0)
                        {
                            cout << "Negative ToT " << tot1[jm][iPlane - 1] << ", for hit= " << jm << endl;
                            cout << "Times1: " << t1t[jm][iPlane - 1][iBar - 1] << " " << t1l[jm][iPlane - 1][iBar - 1]
                                 << endl;
                        }

                        tot2[jm][iPlane - 1][iBar - 1] = t2t[jm][iPlane - 1][iBar - 1] - t2l[jm][iPlane - 1][iBar - 1];
                        if (tot2[jm][iPlane - 1][iBar - 1] < 0)
                        {
                            cout << "Negative ToT " << tot2[jm][iPlane - 1][iBar - 1] << ", for hit= " << jm << endl;
                            cout << "Times2: " << t2t[jm][iPlane - 1][iBar - 1] << " " << t2l[jm][iPlane - 1][iBar - 1]
                                 << endl;
                        }

                        fh_tofd_TotPm[iPlane - 1]->Fill(iBar, tot2[jm][iPlane - 1][iBar - 1]);
                        fh_tofd_TotPm[iPlane - 1]->Fill(-iBar - 1, tot1[jm][iPlane - 1][iBar - 1]);
                    }
                }
    }

    //----------------------------------------------------------------------
    // PTOF
    //----------------------------------------------------------------------

    if (fCalItems.at(DET_PTOF))
    {
        auto det = fCalItems.at(DET_PTOF);

        Double_t tot1 = 0.;
        Double_t tot2 = 0.;
        Double_t t1l = 0.;
        Double_t t2l = 0.;
        Double_t t1t = 0.;
        Double_t t2t = 0.;
        Bool_t bar_quer1 = false;
        Bool_t bar_quer2 = false;

        Int_t nHits = det->GetEntriesFast();
        LOG(DEBUG) << "nHits: " << nHits;
        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BPaddleCalData* hit = (R3BPaddleCalData*)det->At(ihit);

            if (!hit)
                continue; // should not happen

            Int_t iBar = hit->GetBar(); // 1..n
            // get all times of one bar
            t1l = hit->fTime1L_ns;
            t2l = hit->fTime2L_ns;
            t1t = hit->fTime1T_ns;
            t2t = hit->fTime2T_ns;

            if (!(t1l > 0 && t2l > 0 && t1t > 0 && t2t > 0))
                continue;

            fh_ptof_channels->Fill(iBar);
            LOG(DEBUG) << "Bar: " << iBar;
            LOG(DEBUG) << "times PM1: " << t1l << "  " << t1t << "  " << t1t - t1l;
            LOG(DEBUG) << "times PM2: " << t2l << "  " << t2t << "  " << t2t - t2l;
            if (iBar == 7)
                bar_quer1 = true;
            if (iBar == 8)
                bar_quer2 = true;

            // calculate time over threshold and check if clock counter went out of range
            while (t1t - t1l < 0.)
            {
                t1t = t1t + 2048. * fClockFreq;
            }

            while (t2t - t2l < 0.)
            {
                t2t = t2t + 2048. * fClockFreq;
            }
            while (t1l - timeLos[0][0] < 0.)
            {
                t1t = t1t + 2048. * fClockFreq;
                t1l = t1l + 2048. * fClockFreq;
                t2t = t2t + 2048. * fClockFreq;
                t2l = t2l + 2048. * fClockFreq;
            }

            tot1 = t1t - t1l;
            if (tot1 < 0)
            {
                LOG(WARNING) << "Negative ToT " << tot1;
                LOG(WARNING) << "times1: " << t1t << " " << t1l;
            }

            tot2 = t2t - t2l;
            if (tot2 < 0)
            {
                LOG(WARNING) << "Negative ToT " << tot2;
                LOG(WARNING) << "times2: " << t2t << " " << t2l;
            }

            fh_ptof_TotPm1[iBar]->Fill(tot1);
            fh_ptof_TotPm2[iBar]->Fill(tot2);
            if (iBar == 2)
                fh_ptof_test1->Fill(sqrt(tot1 * tot1));
        }

        // once again

        nHits = det->GetEntriesFast();
        //		LOG(DEBUG) << "nHits: " << nHits;
        for (Int_t ihit = 0; ihit < nHits; ihit++)
        {
            R3BPaddleCalData* hit = (R3BPaddleCalData*)det->At(ihit);

            if (!hit)
                continue; // should not happen

            Int_t iBar = hit->GetBar(); // 1..n
            // get all times of one bar
            t1l = hit->fTime1L_ns;
            t2l = hit->fTime2L_ns;
            t1t = hit->fTime1T_ns;
            t2t = hit->fTime2T_ns;

            if (!(t1l > 0 && t2l > 0 && t1t > 0 && t2t > 0))
                continue;

            // calculate time over threshold and check if clock counter went out of range
            while (t1t - t1l < 0.)
            {
                t1t = t1t + 2048. * fClockFreq;
            }

            while (t2t - t2l < 0.)
            {
                t2t = t2t + 2048. * fClockFreq;
            }
            while (t1l - timeLos[0][0] < 0.)
            {
                t1t = t1t + 2048. * fClockFreq;
                t1l = t1l + 2048. * fClockFreq;
                t2t = t2t + 2048. * fClockFreq;
                t2l = t2l + 2048. * fClockFreq;
            }

            tot1 = t1t - t1l;
            if (tot1 < 0)
            {
                LOG(WARNING) << "Negative ToT " << tot1;
                LOG(WARNING) << "times1: " << t1t << " " << t1l;
            }

            tot2 = t2t - t2l;
            if (tot2 < 0)
            {
                LOG(WARNING) << "Negative ToT " << tot2;
                LOG(WARNING) << "times2: " << t2t << " " << t2l;
            }

            if (bar_quer1 && bar_quer2)
            {
                //				fh_ptof_TotPm1[iBar]->Fill(tot1);
                //				fh_ptof_TotPm2[iBar]->Fill(tot2);
                //				if(iBar==2) fh_ptof_test2->Fill(sqrt(tot1*tot2));
                fh_ptof_channels_cut->Fill(iBar);
            }
        }
    }

    myFile.close();

    fNEvents += 1;
}

void R3BOnlineSpectraLosTest::FinishEvent()
{

    for (Int_t det = 0; det < DET_MAX; det++)
    {
        if (fMappedItems.at(det))
        {
            fMappedItems.at(det)->Clear();
        }
        if (fCalItems.at(det))
        {
            fCalItems.at(det)->Clear();
        }
        if (fHitItems.at(det))
        {
            fHitItems.at(det)->Clear();
        }
    }
}

void R3BOnlineSpectraLosTest::FinishTask()
{

    if (fMappedItems.at(DET_ROLU))
    {
        fh_rolu_channels->Write();
        fh_rolu_tot->Write();
    }

    if (fMappedItems.at(DET_LOS))
    {
        if (fNofLosDetectors > 1)
        {
            fh_los_ToT_vs_ToT->Write();
            fh_los_nPart_vs_nPart->Write();
            fh_dt_los1_los2->Write();
        }
        for (Int_t iloscount = 0; iloscount < fNofLosDetectors; iloscount++)
        {
            fh_los_channels[iloscount]->Write();
            fh_los_tot[iloscount]->Write();
            fh_los_dt_hits_ToT[iloscount]->Write();
            fh_los_ihit_ToT[iloscount]->Write();
            fh_los_twin[iloscount]->Write();
            fh_los_tot_mean_corr[iloscount]->Write();
            fh_los_tot_mean[iloscount]->Write();
            fh_los_rw[iloscount]->Write();
            fh_los_tres_MCFD[iloscount]->Write();
            fh_los_tres_TAMEX[iloscount]->Write();
            fh_los_pos_MCFD[iloscount]->Write();
            fh_los_pos_TAMEX[iloscount]->Write();
            fh_los_pos_ToT[iloscount]->Write();
            fh_los_tot_S8[iloscount]->Write();
            fh_los_vftx_tamex[iloscount]->Write();
            // fh_los_tres_MTDC->Write();
        }
        for (Int_t ipm = 0; ipm < 8; ipm++)
        {
            fh_test[ipm]->Write();
            fh_test_corr[ipm]->Write();
        }
    }

    if (fMappedItems.at(DET_BMON))
    {
        fhTpat->Write();
        fh_spill_length->Write();
        fhTrigger->Write();
        fh_IC->Write();
        fh_SEE->Write();
        fh_TOFDOR->Write();
        fh_IC_spill->Write();
        fh_SEE_spill->Write();
        fh_TOFDOR_spill->Write();
    }

    if (fCalItems.at(DET_TOFD))
    {
        fh_TimePreviousEvent->Write();

        for (Int_t i; i < 4; i++)
        {
            fh_tofd_TotPm[i]->Write();
        }
    }
    for (Int_t ifibcount = 0; ifibcount < NOF_FIB_DET; ifibcount++)
    {
        if (fMappedItems.at(ifibcount + DET_FI_FIRST))
        {
            fh_channels_Fib[ifibcount]->Write();
            fh_fibers_Fib[ifibcount]->Write();
            fh_mult_Fib[ifibcount]->Write();
            fh_time_Fib[ifibcount]->Write();
            fh_multihit_m_Fib[ifibcount]->Write();
            fh_multihit_s_Fib[ifibcount]->Write();
            fh_ToT_m_Fib[ifibcount]->Write();
            fh_ToT_s_Fib[ifibcount]->Write();
            fh_ToT_single_Fib[ifibcount]->Write();
            fh_Fib_vs_Events[ifibcount]->Write();
            fh_Fib_ToF[ifibcount]->Write();
            fh_channels_single_Fib[ifibcount]->Write();
        }
    }

    if (fMappedItems.at(DET_PSPX))
    {
        // LOG(INFO) << "Finish MappedPspx";

        for (UInt_t i = 0; i < N_PSPX; i++)
        {
            fh_pspx_channel_x[i]->Write();
            fh_pspx_channel_y[i]->Write();
            fh_pspx_multiplicity_x[i]->Write();
            fh_pspx_multiplicity_y[i]->Write();
            fh_pspx_strips_position[i]->Write();
        }
    }
    if (fCalItems.at(DET_PSPX))
    {
        for (UInt_t i = 0; i < N_PSPX; i++)
        {
            fh_pspx_cal_energy_frontback[i]->Write();
        }
    }
    if (fHitItems.at(DET_PSPX))
    {
        for (UInt_t i = 0; i < (N_PSPX + 1) / 2; i++)
        {
            fh_pspx_hit_energy[i]->Write();
            fh_pspx_hit_position[i]->Write();
        }
    }
}

ClassImp(R3BOnlineSpectraLosTest)
